"""
Module to provide transformations for containers.
"""

import logging
from dataclasses import dataclass
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-lines


@dataclass
class MarkdownChangeRecord:
    """
    Class to keep track of changes.
    """

    is_container_start: bool
    item_b: int
    item_c: MarkdownToken
    item_d: Optional[EndMarkdownToken]


# pylint: disable=too-few-public-methods
class TransformContainers:
    """
    Class to provide transformations for containers.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def handle_current_token(
        current_token: MarkdownToken,
        transformed_data: str,
        container_stack: List[MarkdownToken],
        container_records: List[MarkdownChangeRecord],
        transformed_data_length_before_add: int,
        actual_tokens: List[MarkdownToken],
    ) -> str:
        """
        Handle the current token as far as it concerns any containers.
        """
        if (
            current_token.is_block_quote_start
            or current_token.is_list_start
            or current_token.is_new_list_item
        ):
            TransformContainers.__transform_container_start(
                container_stack,
                container_records,
                transformed_data_length_before_add,
                current_token,
            )
        elif current_token.is_block_quote_end or current_token.is_list_end:
            transformed_data = TransformContainers.__transform_container_end(
                container_stack,
                container_records,
                current_token,
                transformed_data,
                actual_tokens,
            )
        return transformed_data

    # pylint: enable=too-many-arguments

    @staticmethod
    def __transform_container_start(
        container_stack: List[MarkdownToken],
        container_records: List[MarkdownChangeRecord],
        transformed_data_length_before_add: int,
        current_token: MarkdownToken,
    ) -> None:
        if not container_stack:
            container_records.clear()
        container_stack.append(current_token)
        container_records.append(
            MarkdownChangeRecord(
                True, transformed_data_length_before_add, current_token, None
            )
        )
        # POGGER.debug("START:" + ParserHelper.make_value_visible(current_token))
        # POGGER.debug(">>" + ParserHelper.make_value_visible(container_stack))
        # POGGER.debug(">>" + ParserHelper.make_value_visible(container_records))

    @staticmethod
    def __transform_container_end(
        container_stack: List[MarkdownToken],
        container_records: List[MarkdownChangeRecord],
        current_token: MarkdownToken,
        transformed_data: str,
        actual_tokens: List[MarkdownToken],
    ) -> str:
        current_end_token = cast(EndMarkdownToken, current_token)
        POGGER.debug(
            f"END:{ParserHelper.make_value_visible(current_end_token.start_markdown_token)}"
        )
        while container_stack and container_stack[-1].is_new_list_item:
            container_stack.pop()

        MarkdownToken.assert_tokens_are_same_except_for_line_number(
            container_stack.pop(), current_end_token.start_markdown_token
        )
        container_records.append(
            MarkdownChangeRecord(
                False,
                len(transformed_data),
                current_end_token.start_markdown_token,
                current_end_token,
            )
        )
        POGGER.debug(f">>{ParserHelper.make_value_visible(container_stack)}")
        POGGER.debug(f">>{ParserHelper.make_value_visible(container_records)}")

        if not container_stack:
            record_item = container_records[0]
            start_index = record_item.item_b
            adjusted_text = TransformContainers.__apply_container_transformation(
                transformed_data[start_index:], container_records, actual_tokens, []
            )
            POGGER.debug(f"pre>:{transformed_data[: start_index]}:<")
            POGGER.debug(f"adj>:{adjusted_text}:<")
            transformed_data = transformed_data[:start_index] + adjusted_text
            POGGER.debug(f"trn>:{transformed_data}:<")
        return transformed_data

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __apply_line_transformation(
        did_move_ahead: bool,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        current_changed_record: Optional[MarkdownChangeRecord],
        container_line: str,
        actual_tokens: List[MarkdownToken],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        base_line_number: int,
        delta_line: int,
        is_in_multiline_paragraph: bool,
        ort: List[MarkdownToken],
    ) -> str:

        container_line_old = container_line
        token_stack_copy = token_stack[:]
        removed_tokens_copy = removed_tokens[:]
        container_token_indices_copy = container_token_indices[:]
        removed_token_indices_copy = removed_token_indices[:]
        (
            last_container_token_index,
            applied_leading_spaces_to_start_of_container_line,
            container_line,
            did_adjust_due_to_block_quote_start,
        ) = TransformContainers.__apply_primary_transformation(
            did_move_ahead,
            token_stack,
            container_token_indices,
            current_changed_record,
            container_line,
            actual_tokens,
        )

        container_line, block_me = TransformContainers.__adjust_for_list(
            token_stack,
            applied_leading_spaces_to_start_of_container_line,
            container_token_indices,
            container_line,
            container_line_old,
            removed_tokens,
            removed_token_indices,
            current_changed_record,
            ort,
        )
        container_line = TransformContainers.__adjust_for_block_quote(
            token_stack,
            applied_leading_spaces_to_start_of_container_line,
            container_line,
            container_token_indices,
            base_line_number + delta_line,
            did_adjust_due_to_block_quote_start,
            is_in_multiline_paragraph,
            removed_tokens,
            removed_token_indices,
            container_line_old,
        )

        TransformContainers.__adjust_state_for_element(
            token_stack,
            container_token_indices,
            did_move_ahead,
            current_changed_record,
            last_container_token_index,
            removed_tokens,
            removed_token_indices,
            block_me,
        )

        container_token_indices_length = len(container_token_indices)
        if len(container_token_indices_copy) > container_token_indices_length:
            assert (
                len(container_token_indices_copy) == container_token_indices_length + 1
            )
            assert token_stack_copy[-1].is_new_list_item
        for next_index in range(container_token_indices_length):
            if (
                (
                    container_token_indices[next_index]
                    - container_token_indices_copy[next_index]
                )
            ) > 1:
                container_token_indices[next_index] = (
                    container_token_indices_copy[next_index] + 1
                )

        TransformContainers.__apply_line_transformation_check(
            container_line_old,
            container_line,
            removed_token_indices,
            container_token_indices,
            token_stack_copy,
            container_token_indices_copy,
            removed_tokens_copy,
            removed_token_indices_copy,
        )
        return container_line

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __apply_line_transformation_check_loop(
        token_stack_copy: List[MarkdownToken],
        search_index: int,
        container_token_indices_copy: List[int],
        prefix_text_parts: List[str],
        container_line_old: str,
    ) -> str:
        stack_token_copy = token_stack_copy[search_index]
        stack_token_copy_spaces = (
            cast(BlockQuoteMarkdownToken, stack_token_copy).bleading_spaces
            if stack_token_copy.is_block_quote_start
            else cast(ListStartMarkdownToken, stack_token_copy).leading_spaces
        )
        assert stack_token_copy_spaces is not None
        indent_text = stack_token_copy_spaces.split("\n")[
            container_token_indices_copy[search_index]
        ]
        if (
            indent_text
            and indent_text[-1] == ParserLogger.blah_sequence
            and len(prefix_text_parts) == 1
            and container_line_old.startswith(prefix_text_parts[0])
        ):
            container_line_old = container_line_old[len(prefix_text_parts[0]) :]
            prefix_text_parts.insert(0, indent_text[:-1])
        else:
            prefix_text_parts.append(indent_text)
        return container_line_old

    @staticmethod
    def __apply_line_transformation_check_removed(
        removed_tokens_copy: List[MarkdownToken],
        removed_token_indices: List[int],
        removed_token_indices_copy: List[int],
        prefix_text_parts: List[str],
    ) -> None:
        if removed_token_indices[-1] > removed_token_indices_copy[-1]:
            removed_leading_spaces = (
                cast(BlockQuoteMarkdownToken, removed_tokens_copy[-1]).bleading_spaces
                if removed_tokens_copy[-1].is_block_quote_start
                else cast(
                    ListStartMarkdownToken, removed_tokens_copy[-1]
                ).leading_spaces
            )
            assert removed_leading_spaces is not None
            prefix_text_parts.append(
                removed_leading_spaces.split("\n")[removed_token_indices_copy[-1]]
            )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __apply_line_transformation_check(
        container_line_old: str,
        container_line: str,
        removed_token_indices: List[int],
        container_token_indices: List[int],
        token_stack_copy: List[MarkdownToken],
        container_token_indices_copy: List[int],
        removed_tokens_copy: List[MarkdownToken],
        removed_token_indices_copy: List[int],
    ) -> None:
        if container_line_old == container_line or not removed_tokens_copy:
            return
        prefix_text_parts: List[str] = []

        TransformContainers.__apply_line_transformation_check_removed(
            removed_tokens_copy,
            removed_token_indices,
            removed_token_indices_copy,
            prefix_text_parts,
        )
        start_index = len(token_stack_copy) - 1
        if token_stack_copy[start_index].is_new_list_item:
            start_index -= 1
        for search_index in range(start_index, -1, -1):
            if (
                container_token_indices_copy[search_index]
                != container_token_indices[search_index]
            ):
                container_line_old = (
                    TransformContainers.__apply_line_transformation_check_loop(
                        token_stack_copy,
                        search_index,
                        container_token_indices_copy,
                        prefix_text_parts,
                        container_line_old,
                    )
                )

        prefix = "".join(prefix_text_parts[::-1])
        detabified_constructed_line = TabHelper.detabify_string(
            prefix + container_line_old
        )
        detabified_container_line = TabHelper.detabify_string(container_line)
        if detabified_constructed_line != detabified_container_line:
            # TODO: kludge_flag - these SHOULD always match up, but until we do
            # allow for a relief valve
            kludge_flag = True
            assert (
                kludge_flag or detabified_constructed_line == detabified_container_line
            ), f"-->{detabified_constructed_line}=={detabified_container_line}<--"

    # pylint: enable=too-many-arguments

    @staticmethod
    def __abcd_inner(
        removed_tokens: List[MarkdownToken], removed_token_indices: List[int]
    ) -> bool:
        first_token = removed_tokens[0]
        if first_token.is_list_start:
            current_leading_spaces = cast(
                ListStartMarkdownToken, first_token
            ).leading_spaces
            do_check = True
        else:
            bq_token = cast(BlockQuoteMarkdownToken, first_token)
            current_leading_spaces = bq_token.bleading_spaces
            do_check = bq_token.weird_kludge_five
        if do_check:
            split_space_index = (
                len(current_leading_spaces.split("\n"))
                if current_leading_spaces is not None
                else 0
            )
            if split_space_index != removed_token_indices[0]:
                return True
        removed_tokens.pop(0)
        removed_token_indices.pop(0)
        return False

    # pylint: disable=too-many-arguments, too-many-boolean-expressions
    @staticmethod
    def __abcd(
        current_changed_record: Optional[MarkdownChangeRecord],
        actual_tokens: List[MarkdownToken],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        container_stack: List[MarkdownToken],
        container_token_indices: List[int],
        container_line: str,
    ) -> Tuple[Optional[str], str]:  # sourcery skip: hoist-if-from-if
        if (
            len(container_stack) == 1
            and len(removed_tokens) == 3
            and container_stack[0].is_block_quote_start
            and removed_tokens[0].is_block_quote_start
            and removed_tokens[1].is_list_start
            and removed_tokens[2].is_block_quote_start
        ):
            return "", container_line

        assert current_changed_record
        TransformContainers.__xyz(
            current_changed_record, actual_tokens, removed_tokens, removed_token_indices
        )
        prefix = None
        if len(removed_tokens) > 1:
            prefix, container_line = TransformContainers.__abcd_final(
                container_stack,
                container_token_indices,
                removed_tokens,
                removed_token_indices,
                container_line,
            )
            if len(removed_tokens) > 1:
                prefix = TransformContainers.__abcd_remaining(
                    removed_tokens, removed_token_indices, container_line, prefix
                )
        return prefix, container_line

    # pylint: enable=too-many-arguments, too-many-boolean-expressions

    @staticmethod
    def __xyz(
        current_changed_record: MarkdownChangeRecord,
        actual_tokens: List[MarkdownToken],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
    ) -> None:
        token_to_match = (
            current_changed_record.item_c
            if current_changed_record.is_container_start
            else current_changed_record.item_d
        )
        token_index = 0
        while (
            token_index < len(actual_tokens)
            and actual_tokens[token_index] != token_to_match
        ):
            token_index += 1

        while token_index < len(actual_tokens) and (
            actual_tokens[token_index].is_block_quote_end
            or actual_tokens[token_index].is_list_end
        ):
            token_index += 1
        assert token_index != len(actual_tokens)
        while removed_tokens and not TransformContainers.__abcd_inner(
            removed_tokens, removed_token_indices
        ):
            pass

    @staticmethod
    def __abcd_remaining(
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        container_line: str,
        prefix: Optional[str],
    ) -> Optional[str]:
        assert prefix is not None
        while removed_tokens:
            first_removed_token = removed_tokens.pop(0)
            first_removed_index = removed_token_indices.pop(0)

            if first_removed_token.is_block_quote_start:
                bq_token = cast(BlockQuoteMarkdownToken, first_removed_token)
                assert bq_token.bleading_spaces is not None
                current_leading_spaces = bq_token.bleading_spaces.split("\n")
                do_check = bq_token.weird_kludge_five
                x = len(current_leading_spaces) if do_check else 0
            else:
                list_token = cast(ListStartMarkdownToken, first_removed_token)
                assert list_token.leading_spaces is not None
                # if list_token.leading_spaces is None:
                #     x = 0
                # else:
                current_leading_spaces = list_token.leading_spaces.split("\n")
                x = len(current_leading_spaces)
                # endif

            assert x == first_removed_index or (
                first_removed_index + 1 == x
                and not current_leading_spaces[first_removed_index]
            )
            # if x != first_removed_index and (
            #     first_removed_index + 1 != x
            #     or current_leading_spaces[first_removed_index]
            # ):
            #     break
        return "" if container_line.startswith(prefix) else prefix

    # pylint: disable=too-many-arguments
    @staticmethod
    def __abcd_final_list_second_half(
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        container_stack: List[MarkdownToken],
        possible_prefix: Optional[str],
        container_line: str,
        kludge_beta: bool,
    ) -> Optional[str]:
        container_stack_copy: List[MarkdownToken] = container_stack[:]
        container_stack_copy.extend(removed_tokens[::-1])
        is_only_list_starts_on_stack = not any(
            i.is_block_quote_start for i in container_stack_copy
        )

        if kludge_beta:
            return possible_prefix

        if is_only_list_starts_on_stack and possible_prefix is not None:
            for token_index in range(len(removed_tokens) - 1, -1, -1):
                list_token = cast(ListStartMarkdownToken, removed_tokens[token_index])
                assert list_token.leading_spaces is None
                assert removed_token_indices[token_index] == 0

            removed_tokens.clear()
            removed_token_indices.clear()
            current_indent_level = (
                cast(ListStartMarkdownToken, container_stack[-1]).indent_level
                if container_stack
                else 0
            )
            return " " * current_indent_level

        first_removed_token = removed_tokens[0]

        if first_removed_token.is_block_quote_start:
            prefix = TransformContainers.__abcd_final_list_second_half_with_second_block_quote(
                first_removed_token, removed_token_indices[0], possible_prefix
            )
        else:
            prefix = TransformContainers.__abcd_final_list_second_half_with_second_list(
                first_removed_token,
                removed_tokens,
                removed_token_indices,
                possible_prefix,
                container_line,
            )

        removed_tokens.pop(0)
        removed_token_indices.pop(0)
        return prefix

    # pylint: enable=too-many-arguments

    @staticmethod
    def __abcd_final_list_second_half_with_second_block_quote(
        first_removed_token: MarkdownToken,
        first_removed_token_index: int,
        possible_prefix: Optional[str],
    ) -> Optional[str]:
        removed_block_quote_token = cast(BlockQuoteMarkdownToken, first_removed_token)
        assert removed_block_quote_token.bleading_spaces is not None
        split_bleading_spaces = removed_block_quote_token.bleading_spaces.split("\n")
        if first_removed_token_index < len(split_bleading_spaces):
            return split_bleading_spaces[first_removed_token_index] + (
                possible_prefix or ""
            )
        return possible_prefix

    @staticmethod
    def __abcd_final_list_second_half_with_second_list(
        first_removed_token: MarkdownToken,
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        possible_prefix: Optional[str],
        container_line: str,
    ) -> Optional[str]:
        # sourcery skip: equality-identity, inline-variable, move-assign-in-block, remove-assert-true
        assert cast(ListStartMarkdownToken, first_removed_token).leading_spaces is None
        # if removed_list_token.leading_spaces is None:
        # y = 0
        # else:
        #     assert removed_list_token.leading_spaces is not None
        #     y = len(removed_list_token.leading_spaces.split("\n"))
        # endif
        # assert y == 0
        assert not removed_token_indices[0]

        if len(removed_tokens) >= 2:
            removed_tokens.pop(0)
            removed_token_indices.pop(0)
            if removed_tokens[0].is_block_quote_start:
                removed_block_quote_token = cast(
                    BlockQuoteMarkdownToken, removed_tokens[0]
                )
                assert removed_block_quote_token.bleading_spaces is not None
                split_bleading_spaces = removed_block_quote_token.bleading_spaces.split(
                    "\n"
                )
                assert removed_token_indices[0] < len(split_bleading_spaces)
                # if removed_token_indices[0] < len(split_bleading_spaces):
                prefix = f"{split_bleading_spaces[removed_token_indices[0]]}{possible_prefix or ''}"
                return "" if container_line.startswith(prefix) else prefix
                # endif
        return None

    # pylint: disable=too-many-boolean-expressions
    @staticmethod
    def __abcd_final_list(
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        container_stack: List[MarkdownToken],
        container_line: str,
        container_token_indices: List[int],
    ) -> Optional[str]:
        removed_list_token = cast(ListStartMarkdownToken, removed_tokens[0])
        assert removed_list_token.leading_spaces is not None

        split_leading_spaces = removed_list_token.leading_spaces.split("\n")
        # TODO figure out way to do this properly here instead of relying on
        # prefix is None for most cases
        possible_prefix = None
        kludge_beta = False
        if (
            removed_token_indices[0]
            < len(split_leading_spaces)
            # and split_leading_spaces[removed_token_indices[0]]
        ):
            possible_prefix = split_leading_spaces[removed_token_indices[0]]

            if (
                len(removed_tokens) == 2
                and removed_tokens[0].is_list_start
                and removed_tokens[1].is_list_start
                and len(container_stack) == 2
                and container_stack[0].is_block_quote_start
                and container_stack[1].is_list_start
            ):
                block_quote_token = cast(BlockQuoteMarkdownToken, container_stack[0])
                assert block_quote_token.bleading_spaces is not None
                possible_prefix = (
                    block_quote_token.bleading_spaces.split("\n")[
                        container_token_indices[0]
                    ]
                    + possible_prefix
                )
                container_token_indices[0] += 1
                kludge_beta = True

        removed_tokens.pop(0)
        removed_token_indices.pop(0)

        return TransformContainers.__abcd_final_list_second_half(
            removed_tokens,
            removed_token_indices,
            container_stack,
            possible_prefix,
            container_line,
            kludge_beta,
        )

    # pylint: enable=too-many-boolean-expressions

    @staticmethod
    def __abcd_final(
        container_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        container_line: str,
    ) -> Tuple[Optional[str], str]:
        container_stack_copy = container_stack[:]
        container_indices_copy = container_token_indices[:]
        container_stack_copy.extend(removed_tokens[::-1])
        container_indices_copy.extend(removed_token_indices[::-1])

        if container_stack_copy[-1].is_block_quote_start:

            # Temporary until we have more data. test_extra_047f6x
            if (
                len(removed_tokens) == 2
                and container_stack_copy[-2].is_block_quote_start
                and container_stack_copy[-3].is_block_quote_start
            ):
                removed_tokens.clear()
                removed_token_indices.clear()
                return "", container_line

            bq_spaces = cast(
                BlockQuoteMarkdownToken, container_stack_copy[-1]
            ).bleading_spaces
            assert bq_spaces is not None
            bq_split_spaces = bq_spaces.split("\n")
            assert container_indices_copy[-1] == len(bq_split_spaces) - 1
            prefix: str = bq_split_spaces[container_indices_copy[-1]]
            removed_tokens.pop(0)
            removed_token_indices.pop(0)

            if container_stack_copy[-2].is_list_start:
                prefix = TransformContainers.__abcd_final_inner_list(
                    container_stack_copy,
                    container_indices_copy,
                    container_token_indices,
                    removed_tokens,
                    removed_token_indices,
                    prefix,
                )
            else:
                container_line, prefix = (
                    TransformContainers.__abcd_final_list_block_quote(
                        container_stack,
                        container_stack_copy,
                        container_token_indices,
                        container_line,
                        prefix,
                    )
                )
            return prefix, container_line
        prefix_x = TransformContainers.__abcd_final_list(
            removed_tokens,
            removed_token_indices,
            container_stack,
            container_line,
            container_token_indices,
        )
        return prefix_x, container_line

    # pylint: disable=chained-comparison
    @staticmethod
    def __abcd_final_list_block_quote(
        container_stack: List[MarkdownToken],
        container_stack_copy: List[MarkdownToken],
        container_token_indices: List[int],
        container_line: str,
        prefix: str,
    ) -> Tuple[str, str]:
        stack_index = len(container_stack_copy) - 3
        while (
            stack_index >= 0 and container_stack_copy[stack_index].is_block_quote_start
        ):
            stack_index -= 1
        if (
            stack_index >= 0
            and container_stack_copy[stack_index].is_list_start
            and stack_index < len(container_stack)
        ):
            list_token = cast(ListStartMarkdownToken, container_stack[stack_index])
            assert list_token.leading_spaces is not None
            split_spaces = list_token.leading_spaces.split("\n")
            specific_split_space = split_spaces[container_token_indices[stack_index]]
            if specific_split_space.endswith(ParserLogger.blah_sequence):
                assert container_line.startswith(prefix)
                container_line = container_line[len(prefix) :]
                prefix += specific_split_space[:-1]
                container_token_indices[stack_index] += 1
        return container_line, prefix

    # pylint: enable=chained-comparison

    # pylint: disable=too-many-arguments
    @staticmethod
    def __abcd_final_inner_list(
        container_stack_copy: List[MarkdownToken],
        container_indices_copy: List[int],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        prefix: str,
    ) -> str:
        list_spaces = cast(
            ListStartMarkdownToken, container_stack_copy[-2]
        ).leading_spaces
        assert list_spaces is not None
        list_split_spaces = list_spaces.split("\n")
        if container_indices_copy[-2] == len(list_split_spaces) - 1:
            prefix += list_split_spaces[container_indices_copy[-2]]
        else:
            start_index = len(container_indices_copy) - 3
            while (
                start_index >= 0 and not container_stack_copy[start_index].is_list_start
            ):
                start_index -= 1
            assert start_index >= 0
            list_token = cast(ListStartMarkdownToken, container_stack_copy[start_index])
            leading_spaces_index = container_indices_copy[start_index]
            assert list_token.leading_spaces is not None
            prefix += list_token.leading_spaces.split("\n")[leading_spaces_index]

            # if start_index < len(container_stack_copy):
            container_token_indices[start_index] += 1
            # endif
        removed_tokens.pop(0)
        removed_token_indices.pop(0)

        return prefix

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-locals
    @staticmethod
    def __apply_container_transformation(
        container_text: str,
        container_records: List[MarkdownChangeRecord],
        actual_tokens: List[MarkdownToken],
        token_stack: List[MarkdownToken],
    ) -> str:
        # POGGER.debug(
        #     f">>incoming>>:{ParserHelper.make_value_visible(container_text)}:<<"
        # )

        # POGGER.debug(
        #     f">>container_records>>{ParserHelper.make_value_visible(container_records)}"
        # )

        token_stack = []
        container_token_indices: List[int] = []
        base_line_number = container_records[0].item_c.line_number
        container_text_index = container_records[0].item_b
        split_container_text = container_text.split(ParserHelper.newline_character)

        transformed_parts = []
        delta_line = 0
        record_index = -1
        current_changed_record = None
        is_in_multiline_paragraph = False
        POGGER.debug(
            ">>split_container_text>>"
            + ParserHelper.make_value_visible(split_container_text)
        )

        for _, container_line in enumerate(split_container_text):  # pragma: no cover
            container_line_length = len(container_line)
            # POGGER.debug(
            #     ParserHelper.newline_character
            #     + str(delta_line)
            #     + "("
            #     + str(base_line_number + delta_line)
            #     + ")>>container_line>>"
            #     + str(container_text_index)
            #     + "-"
            #     + str(container_text_index + container_line_length)
            #     + ":>:"
            #     + ParserHelper.make_value_visible(container_line)
            #     + ":<"
            # )

            is_para_start_in_line = ParserLogger.start_range_sequence in container_line
            is_para_end_in_line = ParserLogger.end_range_sequence in container_line
            (
                record_index,
                did_move_ahead,
                current_changed_record,
                removed_tokens,
                removed_token_indices,
            ) = TransformContainers.__move_to_current_record(
                record_index,
                container_records,
                container_text_index,
                token_stack,
                container_token_indices,
                container_line_length,
            )

            if not container_token_indices:
                transformed_parts.append(container_line)
                break

            container_line = TransformContainers.__apply_container_transformation_inner(
                container_line,
                actual_tokens,
                did_move_ahead,
                token_stack,
                container_token_indices,
                removed_tokens,
                removed_token_indices,
                current_changed_record,
                base_line_number,
                delta_line,
                is_in_multiline_paragraph,
            )

            TransformContainers.__apply_container_transformation_removed(
                removed_tokens, removed_token_indices
            )

            is_in_multiline_paragraph = (
                not is_para_end_in_line
                if is_in_multiline_paragraph
                else is_para_start_in_line and not is_para_end_in_line
            )

            transformed_parts.append(container_line)
            container_text_index += container_line_length + 1
            delta_line += 1

        # POGGER.debug(
        #     "\n<<transformed<<" + ParserHelper.make_value_visible(transformed_parts)
        # )
        return ParserHelper.newline_character.join(transformed_parts)

    # pylint: enable=too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __apply_container_transformation_inner(
        container_line: str,
        actual_tokens: List[MarkdownToken],
        did_move_ahead: bool,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        current_changed_record: Optional[MarkdownChangeRecord],
        base_line_number: int,
        delta_line: int,
        is_in_multiline_paragraph: bool,
    ) -> str:
        prefix_to_use = None
        if len(removed_tokens) > 1 and current_changed_record:
            prefix_to_use, container_line = TransformContainers.__abcd(
                current_changed_record,
                actual_tokens,
                removed_tokens,
                removed_token_indices,
                token_stack,
                container_token_indices,
                container_line,
            )
        return (
            prefix_to_use + container_line
            if prefix_to_use is not None
            else TransformContainers.__apply_line_transformation(
                did_move_ahead,
                token_stack,
                container_token_indices,
                current_changed_record,
                container_line,
                actual_tokens,
                removed_tokens,
                removed_token_indices,
                base_line_number,
                delta_line,
                is_in_multiline_paragraph,
                removed_tokens[:],
            )
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __apply_container_transformation_removed(
        removed_tokens: List[MarkdownToken], removed_token_indices: List[int]
    ) -> None:
        if removed_tokens:
            last_removed_token = removed_tokens[-1]
            last_removed_token_index = removed_token_indices[-1]

            if last_removed_token.is_block_quote_start:
                last_removed_token_leading_spaces = cast(
                    BlockQuoteMarkdownToken, last_removed_token
                ).bleading_spaces
            else:
                last_removed_token_leading_spaces = cast(
                    ListStartMarkdownToken, last_removed_token
                ).leading_spaces
            calc_index = (
                len(last_removed_token_leading_spaces.split("\n"))
                if last_removed_token_leading_spaces is not None
                else 0
            )
            if last_removed_token_index < calc_index and (
                last_removed_token_index + 1 != calc_index
                or (
                    last_removed_token_leading_spaces is not None
                    and last_removed_token_leading_spaces.split("\n")[-1] != ""
                )
            ):
                fred = calc_index > -1
                assert fred

    # pylint: disable=too-many-arguments
    @staticmethod
    def __move_to_current_record(
        old_record_index: int,
        container_records: List[MarkdownChangeRecord],
        container_text_index: int,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        container_line_length: int,
    ) -> Tuple[
        int,
        bool,
        Optional[MarkdownChangeRecord],
        List[MarkdownToken],
        List[int],
    ]:
        record_index = old_record_index
        current_changed_record = None
        did_move_ahead = False

        POGGER.debug(f"({container_text_index})")
        POGGER.debug(
            f"({record_index + 1}):{ParserHelper.make_value_visible(container_records[1])}"
        )
        while record_index + 1 < len(container_records) and container_records[
            record_index + 1
        ].item_b <= (container_text_index + container_line_length):
            record_index += 1
        POGGER.debug(
            f"({record_index + 1}):{ParserHelper.make_value_visible(container_records[1])}"
        )
        removed_token_indices: List[int] = []
        removed_tokens: List[MarkdownToken] = []
        while old_record_index != record_index:
            (
                old_record_index,
                did_move_ahead,
                current_changed_record,
            ) = TransformContainers.__manage_records(
                container_records,
                old_record_index,
                token_stack,
                container_token_indices,
                removed_tokens,
                removed_token_indices,
            )

        POGGER.debug(
            f"   removed_tokens={ParserHelper.make_value_visible(removed_tokens)}"
        )
        return (
            record_index,
            did_move_ahead,
            current_changed_record,
            removed_tokens,
            removed_token_indices,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __manage_records(
        container_records: List[MarkdownChangeRecord],
        old_record_index: int,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
    ) -> Tuple[int, bool, MarkdownChangeRecord]:
        did_move_ahead, current_changed_record = (
            True,
            container_records[old_record_index + 1],
        )
        POGGER.debug(
            f"   current_changed_record({str(old_record_index + 1)})-->{ParserHelper.make_value_visible(current_changed_record)}"
        )
        if current_changed_record.is_container_start:
            token_stack.append(current_changed_record.item_c)
            container_token_indices.append(0)
        else:
            TransformContainers.__manage_records_check(
                token_stack,
                container_token_indices,
                removed_tokens,
                removed_token_indices,
                current_changed_record,
            )

        POGGER.debug(
            f"   -->current_changed_recordx>{ParserHelper.make_value_visible(current_changed_record)}"
        )
        POGGER.debug(f"   -->{ParserHelper.make_value_visible(token_stack)}")
        POGGER.debug(
            f"   -->{ParserHelper.make_value_visible(container_token_indices)}"
        )
        return old_record_index + 1, did_move_ahead, current_changed_record

    # pylint: enable=too-many-arguments

    @staticmethod
    def __manage_records_check(
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        current_changed_record: Optional[MarkdownChangeRecord],
    ) -> None:
        POGGER.debug(f"   -->{ParserHelper.make_value_visible(token_stack)}")
        POGGER.debug(
            f"   -->{ParserHelper.make_value_visible(container_token_indices)}"
        )

        if token_stack[-1].is_new_list_item:
            removed_tokens.append(token_stack.pop())
            container_token_indices.pop()

        assert current_changed_record is not None
        MarkdownToken.assert_tokens_are_same_except_for_line_number(
            current_changed_record.item_c, token_stack[-1]
        )

        top_of_stack_token = token_stack.pop()
        removed_tokens.append(top_of_stack_token)
        top_of_stack_index = container_token_indices.pop()
        removed_token_indices.append(top_of_stack_index)

        if top_of_stack_token.is_block_quote_start:
            top_of_stack_bq_token = cast(BlockQuoteMarkdownToken, top_of_stack_token)
            top_of_stack_leading_spaces = top_of_stack_bq_token.bleading_spaces
        else:
            top_of_stack_list_token = cast(ListStartMarkdownToken, top_of_stack_token)
            top_of_stack_leading_spaces = top_of_stack_list_token.leading_spaces
        top_of_stack_split_leading_spaces = (
            len(top_of_stack_leading_spaces.split("\n"))
            if top_of_stack_leading_spaces is not None
            else 0
        )

        if top_of_stack_index < top_of_stack_split_leading_spaces and (
            top_of_stack_index + 1 != top_of_stack_split_leading_spaces
            or (
                top_of_stack_leading_spaces is not None
                and top_of_stack_leading_spaces.split("\n")[-1] != ""
            )
        ):
            assert top_of_stack_token is not None

    # pylint: disable=too-many-arguments,too-many-boolean-expressions
    @staticmethod
    def __adjust_state_for_element(
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        did_move_ahead: bool,
        current_changed_record: Optional[MarkdownChangeRecord],
        last_container_token_index: int,
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        block_me: bool,
    ) -> None:
        # POGGER.debug(f" -->{ParserHelper.make_value_visible(token_stack)}")
        # POGGER.debug(f" -->{ParserHelper.make_value_visible(container_token_indices)}")
        did_change_to_list_token = (
            did_move_ahead
            and current_changed_record is not None
            and current_changed_record.is_container_start
            and (token_stack[-1].is_list_start or token_stack[-1].is_new_list_item)
        )

        # Attempt to address one of the open issues.
        # if False:
        #     xx = did_move_ahead and not did_change_to_list_token and last_container_token_index == 0 and \
        #         current_changed_record and current_changed_record.item_c and current_changed_record.item_c.is_block_quote_start and\
        #         len(token_stack) > 2 and token_stack[-1].is_block_quote_start and token_stack[-2].is_block_quote_start
        #     if xx:
        #         x1 = token_stack[-1]
        #         i1 = container_token_indices[len(token_stack) - 1]
        #         l1 = token_stack[-1].bleading_spaces.split("\n")
        #         c1 = l1[i1]
        #         x2 = token_stack[-2]
        #         i2 = container_token_indices[len(token_stack) - 2]
        #         l2 = token_stack[-2].bleading_spaces.split("\n")
        #         if i2 < len(l2) and token_stack[-1].line_number != token_stack[-2].line_number:
        #             c2 = l2[i2]
        #             if c1 and c2 and c1.startswith(c2):
        #                 assert False

        # This corresponds to __do_block_quote_leading_spaces_adjustments_adjust_bleading
        # and yes... this is a kludge.
        if (
            did_move_ahead
            and not did_change_to_list_token
            and last_container_token_index == 0
            and current_changed_record
            and current_changed_record.item_c
            and current_changed_record.item_c.is_block_quote_start
            and len(token_stack) >= 4
            and token_stack[-1].is_block_quote_start
            and token_stack[-2].is_list_start
            and token_stack[-3].is_list_start
            and token_stack[-4].is_block_quote_start
            and token_stack[-1].line_number != token_stack[-2].line_number
        ):
            container_token_indices[-4] += 1

        # May need earlier if both new item and start of new list on same line
        if not did_change_to_list_token:
            TransformContainers.__adjust_state_for_element_inner(
                token_stack,
                container_token_indices,
                removed_tokens,
                removed_token_indices,
                block_me,
                last_container_token_index,
            )
        elif token_stack[-1].is_new_list_item:
            token_stack.pop()
            container_token_indices.pop()
        POGGER.debug(f" -->{ParserHelper.make_value_visible(token_stack)}")
        POGGER.debug(f" -->{ParserHelper.make_value_visible(container_token_indices)}")

    # pylint: enable=too-many-arguments,too-many-boolean-expressions

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_state_for_element_inner(
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        block_me: bool,
        last_container_token_index: int,
    ) -> None:
        previous_token = None
        if removed_tokens and removed_tokens[-1].is_block_quote_start:
            nested_block_start_index = len(removed_tokens) - 1

            previous_token = removed_tokens[nested_block_start_index]
            assert previous_token.is_block_quote_start
            previous_bq_token = cast(BlockQuoteMarkdownToken, previous_token)
            assert previous_bq_token.bleading_spaces is not None
            inner_token_index = removed_token_indices[-1]
            if inner_token_index >= len(previous_bq_token.bleading_spaces.split("\n")):
                previous_token = None
            else:
                removed_token_indices[-1] += 1
        if previous_token is None:
            if not block_me:
                container_token_indices[-1] = last_container_token_index + 1
            current_block_quote_token_index = len(container_token_indices) - 1
            if token_stack[current_block_quote_token_index].is_block_quote_start:
                TransformContainers.__adjust_state_for_element_inner_block_quote(
                    token_stack,
                    container_token_indices,
                    current_block_quote_token_index,
                )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_state_for_element_inner_block_quote(
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        current_block_quote_token_index: int,
    ) -> None:
        previous_block_quote_token_index = current_block_quote_token_index - 1
        while (
            previous_block_quote_token_index >= 0
            and not token_stack[previous_block_quote_token_index].is_block_quote_start
        ):
            previous_block_quote_token_index -= 1

        if (
            previous_block_quote_token_index >= 0
            and token_stack[previous_block_quote_token_index].line_number
            == token_stack[current_block_quote_token_index].line_number
            and not cast(
                BlockQuoteMarkdownToken, token_stack[current_block_quote_token_index]
            ).weird_kludge_three
            and container_token_indices[current_block_quote_token_index] == 1
        ):
            previous_block_quote_token = cast(
                BlockQuoteMarkdownToken, token_stack[previous_block_quote_token_index]
            )
            if not previous_block_quote_token.weird_kludge_seven:
                container_token_indices[previous_block_quote_token_index] += 1
        elif (
            previous_block_quote_token_index >= 0
            and container_token_indices[current_block_quote_token_index] == 1
            and token_stack[previous_block_quote_token_index].line_number
            != token_stack[current_block_quote_token_index].line_number
        ):
            TransformContainers.__adjust_state_for_element_inner_part_1(
                token_stack,
                container_token_indices,
                previous_block_quote_token_index,
                current_block_quote_token_index,
            )

    # pylint: disable=too-many-locals
    @staticmethod
    def __adjust_state_for_element_inner_part_1(
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        previous_block_quote_token_index: int,
        current_block_quote_token_index: int,
    ) -> None:
        rt_previous_token = cast(
            BlockQuoteMarkdownToken, token_stack[previous_block_quote_token_index]
        )
        assert rt_previous_token.bleading_spaces is not None
        rt_previous = rt_previous_token.bleading_spaces.split("\n")
        # rt_current = token_stack[current_block_quote_token_index].bleading_spaces.split(
        #     "\n"
        # )
        # tp_current = rt_current[0]
        ci_prev = container_token_indices[previous_block_quote_token_index]

        current_token = cast(
            BlockQuoteMarkdownToken, token_stack[current_block_quote_token_index]
        )
        if (
            ci_prev < len(rt_previous)
            and token_stack[current_block_quote_token_index].is_block_quote_start
            and current_token.weird_kludge_four is not None
        ):
            # tp_previous = rt_previous[ci_prev]
            fff = current_token.weird_kludge_four
            assert fff is not None

            prev_line = token_stack[previous_block_quote_token_index].line_number
            par_line = fff[0]
            prev_col = token_stack[previous_block_quote_token_index].column_number
            par_col = fff[1]
            prev_cti = container_token_indices[previous_block_quote_token_index]
            par_cti = fff[2]
            prev_leading = rt_previous[prev_cti]
            par_leading = fff[3]

            if (
                prev_line == par_line
                and prev_col == par_col
                and prev_cti == par_cti
                and prev_leading == par_leading
            ):
                container_token_indices[previous_block_quote_token_index] += 1

    # pylint: enable=too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_block_quote(
        token_stack: List[MarkdownToken],
        applied_leading_spaces_to_start_of_container_line: bool,
        container_line: str,
        container_token_indices: List[int],
        line_number: int,
        did_adjust_due_to_block_quote_start: bool,
        is_in_multiline_paragraph: bool,
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        container_line_old: str,
    ) -> str:
        if not (len(token_stack) > 1 and token_stack[-1].is_block_quote_start):
            return container_line

        POGGER.debug(" looking for nested list start")
        nested_list_start_index = TransformContainers.__get_last_list_index_on_stack(
            token_stack
        )
        POGGER.debug(f" afbq={len(token_stack) - 1}")
        POGGER.debug(f" nested_list_start_index={nested_list_start_index}")
        if nested_list_start_index == -1:
            POGGER.debug(" nope")
            return container_line
        if (
            nested_list_start_index == len(token_stack) - 2
            and nested_list_start_index > 0
            and token_stack[-1].line_number == line_number
            and token_stack[nested_list_start_index - 1].is_block_quote_start
            and token_stack[-1].line_number != token_stack[-2].line_number
        ):
            return TransformContainers.__adjust_for_block_quote_same_line(
                container_line,
                nested_list_start_index,
                token_stack,
                container_token_indices,
            )
        return TransformContainers.__adjust_for_block_quote_previous_line(
            container_line,
            nested_list_start_index,
            token_stack,
            container_token_indices,
            line_number,
            applied_leading_spaces_to_start_of_container_line,
            did_adjust_due_to_block_quote_start,
            is_in_multiline_paragraph,
            removed_tokens,
            removed_token_indices,
            container_line_old,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_list(
        token_stack: List[MarkdownToken],
        applied_leading_spaces_to_start_of_container_line: bool,
        container_token_indices: List[int],
        container_line: str,
        container_line_old: str,
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        current_changed_record: Optional[MarkdownChangeRecord],
        ort: List[MarkdownToken],
    ) -> Tuple[str, bool]:
        block_me = False
        if token_stack and (
            token_stack[-1].is_list_start or token_stack[-1].is_new_list_item
        ):
            (
                previous_token,
                inner_token_index,
                nested_block_start_index,
                block_start_on_remove,
                container_line,
                block_me,
            ) = TransformContainers.__adjust_for_list_adjust(
                container_line,
                container_line_old,
                token_stack,
                container_token_indices,
                removed_tokens,
                removed_token_indices,
                applied_leading_spaces_to_start_of_container_line,
            )
            if previous_token:
                container_line, block_me = TransformContainers.__adjust_for_list_end(
                    container_line,
                    token_stack,
                    removed_tokens,
                    removed_token_indices,
                    applied_leading_spaces_to_start_of_container_line,
                    previous_token,
                    container_token_indices,
                    inner_token_index,
                    nested_block_start_index,
                    current_changed_record,
                    block_start_on_remove,
                )
            # elif (
            #     len(ort) == 2
            #     and len(removed_tokens) == 1
            #     and ort[0].is_block_quote_start
            #     and ort[1] == removed_tokens[0]
            #     and ort[1].is_list_start
            # ):
            #     block_me = True
        return container_line, block_me

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_for_list_adjust_block_quote(
        removed_tokens: List[MarkdownToken], removed_token_indices: List[int]
    ) -> Tuple[int, Optional[MarkdownToken], int]:
        nested_block_start_index = len(removed_tokens) - 1
        previous_token: Optional[MarkdownToken] = removed_tokens[
            nested_block_start_index
        ]
        assert previous_token is not None and previous_token.is_block_quote_start

        previous_bq_token = cast(BlockQuoteMarkdownToken, previous_token)
        assert previous_bq_token.bleading_spaces is not None
        inner_token_index = removed_token_indices[-1]
        if inner_token_index >= len(previous_bq_token.bleading_spaces.split("\n")):
            previous_token = None
        return (
            nested_block_start_index,
            previous_token,
            inner_token_index,
        )

    @staticmethod
    def __adjust_for_list_adjust_list(
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        container_line: str,
        container_line_old: str,
    ) -> Tuple[str, bool]:
        block_me = False
        removed_list_token = cast(ListStartMarkdownToken, removed_tokens[-1])
        if removed_list_token.leading_spaces is not None:
            split_list_leading_spaces = removed_list_token.leading_spaces.split("\n")
            removed_index = removed_token_indices[-1]
            if removed_index == 0 and removed_list_token.leading_spaces == "":
                block_me = True
            elif removed_index < len(split_list_leading_spaces):
                if removed_leading_spaces := split_list_leading_spaces[removed_index]:
                    assert container_line.endswith(container_line_old)
                    actual_leading_spaces = container_line[: -len(container_line_old)]

                    block_me = True

                    _, ex_ws = ParserHelper.extract_spaces_verified(container_line, 0)
                    if (
                        "\t" not in ex_ws
                        and not actual_leading_spaces
                        and removed_leading_spaces != actual_leading_spaces
                    ):
                        container_line = removed_leading_spaces + container_line
        return container_line, block_me

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_list_adjust(
        container_line: str,
        container_line_old: str,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        applied_leading_spaces_to_start_of_container_line: bool,
    ) -> Tuple[Optional[MarkdownToken], int, int, bool, str, bool]:
        previous_token = None
        inner_token_index = nested_block_start_index = -1
        block_me = False
        block_start_on_remove = False

        if removed_tokens:
            if removed_tokens[-1].is_block_quote_start:
                block_start_on_remove = True
                (
                    nested_block_start_index,
                    previous_token,
                    inner_token_index,
                ) = TransformContainers.__adjust_for_list_adjust_block_quote(
                    removed_tokens, removed_token_indices
                )
            else:
                assert removed_tokens[-1].is_list_start
                container_line, block_me = (
                    TransformContainers.__adjust_for_list_adjust_list(
                        removed_tokens,
                        removed_token_indices,
                        container_line,
                        container_line_old,
                    )
                )
        if previous_token is None:
            block_start_on_remove = False
            nested_block_start_index = (
                TransformContainers.__find_last_block_quote_on_stack(token_stack)
            )
            if nested_block_start_index != -1:
                POGGER.debug(f"nested_block_start_index>{nested_block_start_index}")
                previous_token = token_stack[nested_block_start_index]
                POGGER.debug(
                    f"previous={ParserHelper.make_value_visible(previous_token)}"
                )
                POGGER.debug(
                    f" applied_leading_spaces_to_start_of_container_line->{applied_leading_spaces_to_start_of_container_line}"
                )
                inner_token_index = container_token_indices[nested_block_start_index]
                # POGGER.debug(
                #     f"applied:{applied_leading_spaces_to_start_of_container_line} or "
                #     + f"end.line:{token_stack[-1].line_number} != prev.line:{previous_token.line_number}"
                # )
        return (
            previous_token,
            inner_token_index,
            nested_block_start_index,
            block_start_on_remove,
            container_line,
            block_me,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_list_end(
        container_line: str,
        token_stack: List[MarkdownToken],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        applied_leading_spaces_to_start_of_container_line: bool,
        previous_token: MarkdownToken,
        container_token_indices: List[int],
        inner_token_index: int,
        nested_block_start_index: int,
        current_changed_record: Optional[MarkdownChangeRecord],
        block_start_on_remove: bool,
    ) -> Tuple[str, bool]:
        block_me = False
        if TransformContainers.__adjust_for_list_check(
            token_stack,
            removed_tokens,
            applied_leading_spaces_to_start_of_container_line,
            previous_token,
            container_line,
        ):
            container_line = TransformContainers.__adjust_for_list_end_part_2(
                container_line,
                previous_token,
                inner_token_index,
                current_changed_record,
            )

        check_end_data = current_changed_record is not None and (
            current_changed_record.item_d is None
            or (
                current_changed_record.item_d is not None
                and not current_changed_record.item_d.extra_end_data
            )
        )
        if (
            not removed_tokens
            or not removed_tokens[-1].is_block_quote_start
            or check_end_data
        ):
            TransformContainers.__adjust_for_list_end_part_3(
                block_start_on_remove,
                token_stack,
                container_token_indices,
                removed_token_indices,
                nested_block_start_index,
                inner_token_index,
            )

        if (
            removed_tokens
            and current_changed_record
            and current_changed_record.item_d is not None
        ):
            container_line, block_me = TransformContainers.__adjust_for_list_end_part_4(
                container_line,
                block_me,
                token_stack,
                container_token_indices,
                removed_tokens,
                removed_token_indices,
            )
        return container_line, block_me

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_for_list_end_part_2(
        container_line: str,
        previous_token: MarkdownToken,
        inner_token_index: int,
        current_changed_record: Optional[MarkdownChangeRecord],
    ) -> str:
        previous_block_token = cast(BlockQuoteMarkdownToken, previous_token)
        assert (
            previous_block_token.bleading_spaces is not None
        ), "Bleading spaces must be defined by this point."
        split_leading_spaces = previous_block_token.bleading_spaces.split(
            ParserHelper.newline_character
        )
        POGGER.debug(
            f"inner_token_index={inner_token_index} < len(split)={len(split_leading_spaces)}"
        )
        assert inner_token_index < len(split_leading_spaces)
        POGGER.debug(
            f" adj-->container_line>:{ParserHelper.make_value_visible(container_line)}:<"
        )

        token_leading_spaces = split_leading_spaces[inner_token_index]
        if (
            current_changed_record
            and not current_changed_record.is_container_start
            and current_changed_record.item_d is not None
            and current_changed_record.item_d.extra_end_data is not None
        ):
            token_end_data = current_changed_record.item_d.extra_end_data
            assert token_end_data.startswith(
                token_leading_spaces
            ) and container_line.startswith(token_end_data)
            token_leading_spaces = ""

        container_line = token_leading_spaces + container_line
        POGGER.debug(
            f" adj-->container_line>:{ParserHelper.make_value_visible(container_line)}:<"
        )
        return container_line

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_list_end_part_3(
        block_start_on_remove: bool,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_token_indices: List[int],
        nested_block_start_index: int,
        inner_token_index: int,
    ) -> None:
        if block_start_on_remove:
            old_index_value = removed_token_indices[nested_block_start_index]
            removed_token_indices[nested_block_start_index] = inner_token_index + 1
        else:
            old_index_value = container_token_indices[nested_block_start_index]
            container_token_indices[nested_block_start_index] = inner_token_index + 1
        ## This is a guess...
        if (
            not block_start_on_remove
            and not old_index_value
            and nested_block_start_index
        ):
            assert token_stack[nested_block_start_index].is_block_quote_start
            block_quote_token = cast(
                BlockQuoteMarkdownToken, token_stack[nested_block_start_index]
            )

            if not block_quote_token.weird_kludge_three:
                for new_start_index in range(nested_block_start_index - 1, -1, -1):
                    if token_stack[new_start_index].is_block_quote_start:
                        container_token_indices[new_start_index] += 1
                        break

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_list_end_part_4(
        container_line: str,
        block_me: bool,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
    ) -> Tuple[str, bool]:
        if token_stack[-1].is_list_start and removed_tokens[-1].is_list_start:
            removed_list_token = cast(ListStartMarkdownToken, removed_tokens[-1])
            assert removed_list_token.leading_spaces is not None
            # if removed_list_token.leading_spaces is None:
            #     assert removed_token_index == 0
            # else:
            removed_token_split_spaces = removed_list_token.leading_spaces.split("\n")
            removed_token_index = removed_token_indices[-1]
            assert removed_token_index < len(removed_token_split_spaces)
            removed_token_indices[-1] += 1
            # end else
            return container_line, True
        if removed_tokens[-1].is_block_quote_start and token_stack[-1].is_list_start:
            container_line = TransformContainers.__adjust_for_list_end_part_4_inner(
                container_line,
                token_stack,
                container_token_indices,
                removed_tokens,
                removed_token_indices,
            )
        container_token_indices[-1] += 1
        return container_line, block_me

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_for_list_end_part_4_inner(
        container_line: str,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
    ) -> str:
        removed_block_quote_token = cast(BlockQuoteMarkdownToken, removed_tokens[-1])
        assert removed_block_quote_token.bleading_spaces is not None
        list_token = cast(ListStartMarkdownToken, token_stack[-1])
        assert list_token.leading_spaces is not None
        split_leading_spaces = list_token.leading_spaces.split("\n")
        removed_split_leading_spaces = removed_block_quote_token.bleading_spaces.split(
            "\n"
        )
        removed_token_index = removed_token_indices[-1]

        if removed_token_index < len(removed_split_leading_spaces):
            leading_space_to_use = removed_split_leading_spaces[removed_token_index]
            leading_index_to_use = split_leading_spaces[container_token_indices[-1]]
            # if leading_index_to_use.endswith(
            #     ParserLogger.blah_sequence
            # ) and container_line.startswith(leading_space_to_use):
            assert leading_index_to_use.endswith(
                ParserLogger.blah_sequence
            ) and container_line.startswith(leading_space_to_use)
            container_line = f"{leading_space_to_use}{leading_index_to_use[:-1]}{container_line[len(leading_space_to_use) :]}"
            # end if
        return container_line

    @staticmethod
    def __adjust_for_list_check(
        token_stack: List[MarkdownToken],
        removed_tokens: List[MarkdownToken],
        applied_leading_spaces_to_start_of_container_line: bool,
        previous_token: MarkdownToken,
        container_line: str,
    ) -> bool:
        if not token_stack[-1].is_new_list_item:
            return (
                applied_leading_spaces_to_start_of_container_line
                or token_stack[-1].line_number != previous_token.line_number
            )
        new_list_item_adjust = True
        if len(removed_tokens) == 1 and removed_tokens[-1].is_block_quote_start:
            removed_block_token = cast(BlockQuoteMarkdownToken, removed_tokens[-1])
            assert (
                removed_block_token.bleading_spaces is not None
            ), "Bleading spaces must be defined by this point."
            leading_spaces_newline_count = removed_block_token.bleading_spaces.count(
                "\n"
            )
            # block_quote_end_line = (
            #     leading_spaces_newline_count + removed_block_token.line_number
            # )
            # POGGER.debug(
            #     f"block_quote_end_line={block_quote_end_line} = "
            #     + f"fg={leading_spaces_newline_count} + "
            #     + f"line={removed_block_token.line_number}"
            # )
            new_list_item_adjust = leading_spaces_newline_count > 1 and (
                removed_block_token.weird_kludge_one is None
                or removed_block_token.weird_kludge_one <= 1
            )
            POGGER.debug(f"new_list_item_adjust:{new_list_item_adjust}")

            if new_list_item_adjust and container_line:
                new_list_item_adjust = TransformContainers.__look_for_container_prefix(
                    token_stack, container_line
                )
        return (
            token_stack[-1].line_number != previous_token.line_number
            and new_list_item_adjust
        )

    @staticmethod
    def __look_for_container_prefix(
        token_stack: List[MarkdownToken], container_line: str
    ) -> bool:
        end_stack_index = len(token_stack) - 1
        assert token_stack[end_stack_index].is_new_list_item
        end_stack_index -= 1
        assert token_stack[end_stack_index].is_list_start

        stack_index = 0
        container_lindex_index, _ = ParserHelper.collect_while_spaces_verified(
            container_line, 0
        )
        is_tracking = True
        while stack_index < end_stack_index and is_tracking:
            # was if
            assert token_stack[stack_index].is_block_quote_start
            is_tracking = ParserHelper.is_character_at_index(
                container_line, container_lindex_index, ">"
            )
            assert is_tracking
            # if is_tracking:
            container_lindex_index, _ = ParserHelper.collect_while_spaces_verified(
                container_line, container_lindex_index + 1
            )
            # end if is_tracking
            # endif
            stack_index += 1
        assert is_tracking
        list_token = cast(ListStartMarkdownToken, token_stack[end_stack_index])
        if not list_token.is_unordered_list_start:
            container_lindex_index, numeric_prefix = (
                ParserHelper.collect_while_one_of_characters_verified(
                    container_line, container_lindex_index, "0123456789"
                )
            )
            assert len(numeric_prefix) > 0
        return not ParserHelper.is_character_at_index(
            container_line, container_lindex_index, list_token.list_start_sequence
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_block_quote_previous_line(
        container_line: str,
        nested_list_start_index: int,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        line_number: int,
        applied_leading_spaces_to_start_of_container_line: bool,
        did_adjust_due_to_block_quote_start: bool,
        is_in_multiline_paragraph: bool,
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
        container_line_old: str,
    ) -> str:
        previous_cl = container_line
        previous_token = token_stack[nested_list_start_index]
        # POGGER.debug(f"nested_list_start_index->{nested_list_start_index}")
        # POGGER.debug(f" yes->{ParserHelper.make_value_visible(previous_token)}")

        # POGGER.debug(f"token_stack[-1].line_number->{token_stack[-1].line_number}")
        # POGGER.debug(f"previous_token.line_number->{previous_token.line_number}")
        # POGGER.debug(f"line_number->{line_number}")
        if (
            token_stack[-1].line_number != previous_token.line_number
            or line_number != previous_token.line_number
        ):
            container_line = TransformContainers.__adjust_for_block_quote_previous_line_nudge_different(
                container_line,
                nested_list_start_index,
                token_stack,
                container_token_indices,
                did_adjust_due_to_block_quote_start,
                is_in_multiline_paragraph,
                applied_leading_spaces_to_start_of_container_line,
                removed_tokens,
                removed_token_indices,
            )
        else:
            POGGER.debug("same line as list start")
            container_line = (
                TransformContainers.__adjust_for_block_quote_previous_line_nudge_same(
                    container_line,
                    nested_list_start_index,
                    token_stack,
                    container_token_indices,
                    previous_token,
                )
            )
        container_line = (
            TransformContainers.__adjust_for_block_quote_previous_line_nudge(
                container_line, previous_cl, container_line_old
            )
        )
        return container_line

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_block_quote_previous_line_nudge_different(
        container_line: str,
        nested_list_start_index: int,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        did_adjust_due_to_block_quote_start: bool,
        is_in_multiline_paragraph: bool,
        applied_leading_spaces_to_start_of_container_line: bool,
        removed_tokens: List[MarkdownToken],
        removed_token_indices: List[int],
    ) -> str:
        POGGER.debug("different line as list start")
        is_special_case = not (
            did_adjust_due_to_block_quote_start and not is_in_multiline_paragraph
        )
        container_line_change_required = (
            not applied_leading_spaces_to_start_of_container_line or is_special_case
        )
        if removed_tokens and removed_tokens[0].is_list_start:
            removed_list_token = cast(ListStartMarkdownToken, removed_tokens[0])
            if removed_list_token.leading_spaces is not None and removed_token_indices[
                0
            ] < len(removed_list_token.leading_spaces.split("\n")):
                nested_list_start_index = 0
                token_stack = removed_tokens
                container_token_indices = removed_token_indices
        return TransformContainers.__adjust(
            nested_list_start_index,
            token_stack,
            container_token_indices,
            container_line,
            False,
            apply_change_to_container_line=container_line_change_required,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_for_block_quote_previous_line_nudge_same(
        container_line: str,
        nested_list_start_index: int,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        previous_token: MarkdownToken,
    ) -> str:
        if nested_list_start_index > 0:
            pre_previous_token = token_stack[nested_list_start_index - 1]
            # POGGER.debug(
            #     f" pre_previous_token->{ParserHelper.make_value_visible(pre_previous_token)}"
            # )
            if pre_previous_token.is_block_quote_start:
                different_line_prefix = TransformContainers.__adjust(
                    nested_list_start_index - 1,
                    token_stack,
                    container_token_indices,
                    "",
                    False,
                )
                # POGGER.debug(f"different_line_prefix>:{different_line_prefix}:<")
                if pre_previous_token.line_number != previous_token.line_number:
                    container_line = different_line_prefix + container_line
        return container_line

    @staticmethod
    def __adjust_for_block_quote_previous_line_nudge(
        container_line: str, previous_cl: str, container_line_old: str
    ) -> str:
        if previous_cl != container_line and container_line.endswith(previous_cl):
            adj_container_line = container_line[: -len(previous_cl)]
            if adj_container_line.endswith(ParserLogger.blah_sequence):
                assert previous_cl.endswith(container_line_old)
                container_line = (
                    previous_cl[: -len(container_line_old)]
                    + adj_container_line[:-1]
                    + previous_cl[-len(container_line_old) :]
                )
        return container_line

    @staticmethod
    def __adjust_for_block_quote_same_line(
        container_line: str,
        nested_list_start_index: int,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
    ) -> str:
        adj_line = TransformContainers.__adjust(
            nested_list_start_index - 1,
            token_stack,
            container_token_indices,
            "",
            True,
        )
        POGGER.debug(f"adj_line->:{adj_line}:")
        adj_line = TransformContainers.__adjust(
            nested_list_start_index,
            token_stack,
            container_token_indices,
            adj_line,
            True,
        )
        POGGER.debug(f"adj_line->:{adj_line}:")
        return adj_line + container_line

    # pylint: disable=too-many-arguments
    @staticmethod
    def __apply_primary_transformation(
        did_move_ahead: bool,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        current_changed_record: Optional[MarkdownChangeRecord],
        container_line: str,
        actual_tokens: List[MarkdownToken],
    ) -> Tuple[int, bool, str, bool]:
        POGGER.debug(
            f" -->did_move_ahead>{ParserHelper.make_value_visible(did_move_ahead)}"
        )
        POGGER.debug(f" -->{ParserHelper.make_value_visible(token_stack)}")
        POGGER.debug(f" -->{ParserHelper.make_value_visible(container_token_indices)}")
        POGGER.debug(
            f" -->current_changed_record>{ParserHelper.make_value_visible(current_changed_record)}"
        )

        is_list_start_after_two_block_starts = (
            TransformContainers.__apply_primary_transformation_start(
                current_changed_record, actual_tokens
            )
        )

        was_abrupt_block_quote_end = bool(
            current_changed_record
            and current_changed_record.item_d is not None
            and current_changed_record.item_d.is_block_quote_end
        )
        if was_abrupt_block_quote_end:
            assert (
                current_changed_record is not None
            ), "If an abrupt bq end, the change record must be defined."
            assert (
                current_changed_record.item_d is not None
            ), "If an abrupt bq end, the change record's item_d field must be defined."
            was_abrupt_block_quote_end = bool(
                current_changed_record.item_d.was_forced
                and current_changed_record.item_d.extra_end_data
                and ">" in current_changed_record.item_d.extra_end_data
            )

        applied_leading_spaces_to_start_of_container_line = (
            (
                not did_move_ahead
                or current_changed_record is None
                or not current_changed_record.is_container_start
            )
            and not is_list_start_after_two_block_starts
            and not was_abrupt_block_quote_end
        )

        last_container_token_index = container_token_indices[-1]
        if applied_leading_spaces_to_start_of_container_line:
            container_line, did_adjust_due_to_block_quote_start = (
                TransformContainers.__apply_primary_transformation_adjust_container_line(
                    token_stack,
                    last_container_token_index,
                    container_line,
                )
            )
        else:
            did_adjust_due_to_block_quote_start = False
        return (
            last_container_token_index,
            applied_leading_spaces_to_start_of_container_line,
            container_line,
            did_adjust_due_to_block_quote_start,
        )

    # pylint: enable=too-many-arguments

    @classmethod
    def __apply_primary_transformation_adjust_container_line(
        cls,
        token_stack: List[MarkdownToken],
        last_container_token_index: int,
        container_line: str,
    ) -> Tuple[str, bool]:
        POGGER.debug(f" container->{ParserHelper.make_value_visible(token_stack[-1])}")
        did_adjust_due_to_block_quote_start = False
        tabbed_leading_space: Optional[str] = None
        if token_stack[-1].is_block_quote_start:
            did_adjust_due_to_block_quote_start = True
            prev_block_token = cast(BlockQuoteMarkdownToken, token_stack[-1])
            assert (
                prev_block_token.bleading_spaces is not None
            ), "Bleading spaces must be defined by this point."
            split_leading_spaces = prev_block_token.bleading_spaces.split(
                ParserHelper.newline_character
            )
            if last_container_token_index in prev_block_token.tabbed_bleading_spaces:
                tabbed_leading_space = prev_block_token.tabbed_bleading_spaces[
                    last_container_token_index
                ]
        else:
            prev_list_token = (
                cast(ListStartMarkdownToken, token_stack[-2])
                if token_stack[-1].is_new_list_item
                else cast(ListStartMarkdownToken, token_stack[-1])
            )
            split_leading_spaces = None
            # following assert was if over next line
            assert prev_list_token.leading_spaces is not None
            split_leading_spaces = prev_list_token.leading_spaces.split(
                ParserHelper.newline_character
            )
            # endif
        if split_leading_spaces is not None and last_container_token_index < len(
            split_leading_spaces
        ):
            POGGER.debug(f" -->{ParserHelper.make_value_visible(split_leading_spaces)}")
            POGGER.debug(
                f" primary-->container_line>:{ParserHelper.make_value_visible(container_line)}:<"
            )
            container_line = (
                tabbed_leading_space + container_line
                if tabbed_leading_space
                else split_leading_spaces[last_container_token_index] + container_line
            )
            POGGER.debug(
                f" -->container_line>:{ParserHelper.make_value_visible(container_line)}:<"
            )
        else:
            did_adjust_due_to_block_quote_start = False
        return container_line, did_adjust_due_to_block_quote_start

    @staticmethod
    def __apply_primary_transformation_start(
        current_changed_record: Optional[MarkdownChangeRecord],
        actual_tokens: List[MarkdownToken],
    ) -> bool:
        if current_changed_record and current_changed_record.item_c.is_list_start:
            list_start_token = current_changed_record.item_c
            list_start_token_index = actual_tokens.index(list_start_token)
            POGGER.debug(
                f" -->list_start_token_index>{ParserHelper.make_value_visible(list_start_token_index)}"
            )

            # pylint: disable=too-many-boolean-expressions
            if (
                list_start_token_index >= 2
                and actual_tokens[list_start_token_index - 1].is_block_quote_start
                and actual_tokens[list_start_token_index - 2].is_block_quote_start
                and actual_tokens[list_start_token_index - 1].line_number
                == list_start_token.line_number
                and actual_tokens[list_start_token_index - 2].line_number
                == list_start_token.line_number
                and actual_tokens[list_start_token_index + 1].is_list_end
                and actual_tokens[list_start_token_index + 2].is_blank_line
            ):
                return True
                # pylint: enable=too-many-boolean-expressions
        return False

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust(
        nested_list_start_index: int,
        token_stack: List[MarkdownToken],
        container_token_indices: List[int],
        container_line: str,
        apply_list_fix: bool,
        apply_change_to_container_line: bool = True,
    ) -> str:
        previous_token = token_stack[nested_list_start_index]
        if (
            apply_list_fix
            and apply_change_to_container_line
            and previous_token.is_list_start
        ):
            previous_list_token = cast(ListStartMarkdownToken, previous_token)
            delta = previous_list_token.indent_level - len(container_line)
            POGGER.debug(f"delta->{delta}")
            container_line += " " * delta

        if previous_token.is_block_quote_start:
            previous_block_token = cast(BlockQuoteMarkdownToken, previous_token)
            leading_spaces = previous_block_token.bleading_spaces or ""
        else:
            if previous_token.is_new_list_item:
                previous_token = token_stack[nested_list_start_index - 1]
                assert previous_token.is_list_start
            previous_list_token = cast(ListStartMarkdownToken, previous_token)
            leading_spaces = previous_list_token.leading_spaces or ""

        split_leading_spaces = leading_spaces.split(ParserHelper.newline_character)
        inner_token_index = container_token_indices[nested_list_start_index]
        assert inner_token_index < len(split_leading_spaces)
        POGGER.debug(
            f"inner_index->{str(container_token_indices[nested_list_start_index])}"
        )
        if apply_change_to_container_line:
            container_line = split_leading_spaces[inner_token_index] + container_line
        container_token_indices[nested_list_start_index] += 1
        POGGER.debug(
            f"inner_index->{str(container_token_indices[nested_list_start_index])}"
        )
        return container_line

    # pylint: enable=too-many-arguments

    @classmethod
    def __get_last_list_index_on_stack(cls, token_stack: List[MarkdownToken]) -> int:
        return next(
            (
                stack_index
                for stack_index in range(len(token_stack) - 2, -1, -1)
                if token_stack[stack_index].is_list_start
                or token_stack[stack_index].is_new_list_item
            ),
            -1,
        )

    @staticmethod
    def __find_last_block_quote_on_stack(token_stack: List[MarkdownToken]) -> int:
        return next(
            (
                stack_index
                for stack_index in range(len(token_stack) - 2, -1, -1)
                if token_stack[stack_index].is_block_quote_start
            ),
            -1,
        )


# pylint: enable=too-few-public-methods



"""
Module to provide structure to extend the parser
"""

from abc import ABC, abstractmethod

from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade


class ParserExtension(ABC):
    """
    Class to provide structure for extensions to how the parser handles various concepts.
    """

    @abstractmethod
    def get_identifier(self) -> str:
        """
        Get the identifier associated with this extension.
        """

    @abstractmethod
    def get_details(self) -> ExtensionDetails:
        """
        Get the details for the extension.
        """

    @abstractmethod
    def apply_configuration(
        self, extension_specific_facade: MyApplicationPropertiesFacade
    ) -> None:
        """
        Apply any configuration required by the extension.
        """



"""
Module to implement a plugin that looks for multiple blank lines in the files.
"""

from dataclasses import dataclass
from typing import Dict, List, Optional, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.utils.leading_space_index_tracker import (
    LeadingSpaceIndexTracker,
)
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken


@dataclass
class PendingContainerFixes:
    """Class to hold the pending fixes to be made when the container is closed."""

    leading_space_index: int
    repeat_count: int


class RuleMd012(RulePlugin):
    """
    Class to implement a plugin that looks for multiple blank lines in the files.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__blank_lines_maximum = 0
        self.__blank_line_count = 0
        self.__last_blank_line: Optional[MarkdownToken] = None
        self.__captured_blank_line: List[MarkdownToken] = []
        self.__leading_space_index_tracker = LeadingSpaceIndexTracker()
        self.__container_fix_map: Dict[MarkdownToken, List[PendingContainerFixes]] = {}

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="no-multiple-blanks",
            plugin_id="MD012",
            plugin_enabled_by_default=True,
            plugin_description="Multiple consecutive blank lines",
            plugin_version="0.7.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md012.md",
            plugin_configuration="maximum",
            plugin_supports_fix=True,
            plugin_fix_level=1,
        )

    @classmethod
    def __validate_maximum(cls, found_value: int) -> None:
        if found_value < 0:
            raise ValueError("Allowable values are any non-negative integers.")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__blank_lines_maximum = (
            self.plugin_configuration.get_integer_property_with_default(
                "maximum",
                1,
                valid_value_fn=self.__validate_maximum,
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [QueryConfigItem("maximum", self.__blank_lines_maximum)]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__blank_line_count = 0
        self.__last_blank_line = None
        self.__captured_blank_line.clear()
        self.__leading_space_index_tracker.clear()
        self.__container_fix_map.clear()

    def __append_to_container_fix_map(
        self, key: MarkdownToken, value: PendingContainerFixes
    ) -> None:
        if key in self.__container_fix_map:
            value_list = self.__container_fix_map[key]
        else:
            value_list = []
            self.__container_fix_map[key] = value_list
        value_list.append(value)

    def __fix_containers(self) -> None:
        outer_container_index = (
            self.__leading_space_index_tracker.get_container_stack_size() - 1
        )
        container_token = self.__leading_space_index_tracker.get_container_stack_item(
            outer_container_index
        )
        line_number_delta = (
            self.__captured_blank_line[-1].line_number
            - self.__captured_blank_line[0].line_number
        )
        if container_token.is_block_quote_start:
            leading_space_index = self.__leading_space_index_tracker.get_tokens_block_quote_bleading_space_index(
                self.__captured_blank_line[-1]
            )
            self.__append_to_container_fix_map(
                container_token,
                PendingContainerFixes(
                    leading_space_index - (line_number_delta - 1), line_number_delta
                ),
            )
            inner_container_index = outer_container_index - 1
            while (
                inner_container_index >= 0
                and self.__leading_space_index_tracker.get_container_stack_item(
                    inner_container_index
                ).is_block_quote_start
            ):
                inner_container_index -= 1
            if inner_container_index >= 0:
                leading_space_index = self.__leading_space_index_tracker.get_tokens_list_leading_space_index(
                    self.__captured_blank_line[-1], inner_container_index
                )
                container_token = (
                    self.__leading_space_index_tracker.get_container_stack_item(
                        inner_container_index
                    )
                )
                self.__append_to_container_fix_map(
                    container_token,
                    PendingContainerFixes(
                        leading_space_index - (line_number_delta - 1), line_number_delta
                    ),
                )
        else:
            leading_space_index = (
                self.__leading_space_index_tracker.get_tokens_list_leading_space_index(
                    self.__captured_blank_line[-1], outer_container_index
                )
            )
            self.__append_to_container_fix_map(
                container_token,
                PendingContainerFixes(
                    leading_space_index - (line_number_delta - 1), line_number_delta
                ),
            )
            inner_container_index = outer_container_index - 1
            while (
                inner_container_index >= 0
                and self.__leading_space_index_tracker.get_container_stack_item(
                    inner_container_index
                ).is_list_start
            ):
                inner_container_index -= 1
            if inner_container_index >= 0:
                leading_space_index = self.__leading_space_index_tracker.get_tokens_block_quote_bleading_space_index(
                    self.__captured_blank_line[-1], inner_container_index
                )
                container_token = (
                    self.__leading_space_index_tracker.get_container_stack_item(
                        inner_container_index
                    )
                )
                self.__append_to_container_fix_map(
                    container_token,
                    PendingContainerFixes(
                        leading_space_index - (line_number_delta - 1), line_number_delta
                    ),
                )

    def __check_for_excess_blank_lines(self, context: PluginScanContext) -> None:
        if self.__blank_line_count > self.__blank_lines_maximum:
            if context.in_fix_mode:
                if self.__leading_space_index_tracker.in_at_least_one_container():
                    self.__fix_containers()

                replacement_tokens = self.__captured_blank_line[:1]
                self.register_replace_tokens_request(
                    context,
                    self.__captured_blank_line[0],
                    self.__captured_blank_line[-1],
                    replacement_tokens,
                )
            else:
                assert self.__last_blank_line is not None
                extra_data = f"Expected: {self.__blank_lines_maximum}, Actual: {self.__blank_line_count}"
                self.report_next_token_error(
                    context, self.__last_blank_line, extra_error_information=extra_data
                )
        self.__captured_blank_line.clear()

    def __process_pending_container_end_fixes(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        end_token = cast(EndMarkdownToken, token)
        if end_token.start_markdown_token in self.__container_fix_map:
            value_list = self.__container_fix_map[end_token.start_markdown_token]
            if end_token.start_markdown_token.is_block_quote_start:
                block_token = cast(
                    BlockQuoteMarkdownToken, end_token.start_markdown_token
                )
                assert block_token.bleading_spaces is not None
                split_spaces = block_token.bleading_spaces.split("\n")
                token_part_name = "bleading_spaces"
            else:
                list_token = cast(
                    ListStartMarkdownToken, end_token.start_markdown_token
                )
                assert list_token.leading_spaces is not None
                split_spaces = list_token.leading_spaces.split("\n")
                token_part_name = "leading_spaces"

            while value_list:
                pending_container_fixes: PendingContainerFixes = value_list[-1]
                for _ in range(pending_container_fixes.repeat_count):
                    del split_spaces[pending_container_fixes.leading_space_index]
                del value_list[-1]

            self.register_fix_token_request(
                context,
                end_token.start_markdown_token,
                "next_token",
                token_part_name,
                "\n".join(split_spaces),
            )

    def completed_file(self, context: PluginScanContext) -> None:
        """
        Event that the file being currently scanned is now completed.
        """
        self.__check_for_excess_blank_lines(context)

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_blank_line:
            if (
                self.__last_blank_line is not None
                and (token.line_number - self.__last_blank_line.line_number) != 1
            ):
                self.__check_for_excess_blank_lines(context)
                self.__blank_line_count = 0
            self.__captured_blank_line.append(token)
            self.__last_blank_line = token
            self.__blank_line_count += 1
        else:
            if self.__blank_line_count:
                self.__check_for_excess_blank_lines(context)
            self.__blank_line_count = 0

        if token.is_block_quote_end or token.is_list_end:
            self.__process_pending_container_end_fixes(context, token)

        if not token.is_end_token or token.is_end_of_stream:
            while (
                self.__leading_space_index_tracker.have_any_registered_container_ends()
            ):
                self.__leading_space_index_tracker.process_container_end(token)
        if token.is_block_quote_start or token.is_list_start:
            self.__leading_space_index_tracker.open_container(token)
        elif token.is_block_quote_end or token.is_list_end:
            self.__leading_space_index_tracker.register_container_end(token)
        elif token.is_new_list_item:
            self.__leading_space_index_tracker.nudge_list_container(token)
        self.__leading_space_index_tracker.track_since_last_non_end_token(token)



"""
Module to implement a plugin that looks for headings that are not surrounded by
blank lines.
"""

from typing import List, Optional

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken

# pylint: disable=too-many-instance-attributes


class RuleMd022(RulePlugin):
    """
    Class to implement a plugin that looks for headings that are not surrounded by
    blank lines.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__blank_line_count = -1
        self.__did_above_line_count_match = False
        self.__start_heading_token: Optional[MarkdownToken] = None
        self.__did_heading_end = False
        self.__lines_above = 0
        self.__lines_below = 0
        self.__start_heading_blank_line_count = -1
        self.__last_blank_line: Optional[MarkdownToken] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="blanks-around-headings,blanks-around-headers",
            plugin_id="MD022",
            plugin_enabled_by_default=True,
            plugin_description="Headings should be surrounded by blank lines.",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md022.md",
            plugin_configuration="lines_above, lines_below",
        )

    @classmethod
    def __validate_configuration_value(cls, found_value: int) -> None:
        if found_value < 0:
            raise ValueError("Value must not be zero or a positive integer.")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__lines_above = (
            self.plugin_configuration.get_integer_property_with_default(
                "lines_above",
                1,
                valid_value_fn=RuleMd022.__validate_configuration_value,
            )
        )
        self.__lines_below = (
            self.plugin_configuration.get_integer_property_with_default(
                "lines_below",
                1,
                valid_value_fn=RuleMd022.__validate_configuration_value,
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("lines_above", self.__lines_above),
            QueryConfigItem("lines_below", self.__lines_below),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__blank_line_count = -1
        self.__did_above_line_count_match = False
        self.__start_heading_token = None
        self.__did_heading_end = False
        self.__last_blank_line = None

    def __next_token_heading_start(self, token: MarkdownToken) -> None:
        # print(">>token.is_setext_heading or token.is_atx_heading>>")
        self.__did_above_line_count_match = self.__blank_line_count in [
            -1,
            self.__lines_above,
        ]

        self.__start_heading_token = token
        self.__start_heading_blank_line_count = self.__blank_line_count
        self.__did_heading_end = False
        # print("self.__did_above_line_count_match>>" + str(self.__did_above_line_count_match))

    def __next_token_heading_end(self, token: MarkdownToken) -> None:
        # print(">>token.is_end_token>>")
        if not token.is_list_end and not token.is_block_quote_end:
            self.__blank_line_count = (
                0 if token.is_leaf_end_token or token.is_container_end_token else -1
            )
        if token.is_atx_heading_end or token.is_setext_heading_end:
            self.__did_heading_end = True

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        # print("START>>" + str(token).replace(ParserHelper.newline_character, "\\n").replace("\t", "\\t"))
        # print(">>self.__blank_line_count>>" + str(self.__blank_line_count))
        if (self.__blank_line_count != -1) and self.__blank_line_count >= 0:
            self.perform_close_check(context, token)
        elif (
            self.__blank_line_count == -1
            and token.is_blank_line
            and self.__last_blank_line is not None
        ):
            blank_line_number_delta = (
                token.line_number - self.__last_blank_line.line_number
            )
            if blank_line_number_delta != 1:
                self.__blank_line_count = 0

        if (
            token.is_blank_line
            and (self.__blank_line_count != -1)
            and self.__blank_line_count >= 0
        ):
            self.__blank_line_count += 1
        if token.is_setext_heading or token.is_atx_heading:
            self.__next_token_heading_start(token)
        elif token.is_thematic_break or token.is_link_reference_definition:
            self.__blank_line_count = 0
        elif token.is_end_token:
            self.__next_token_heading_end(token)
        # print(">>self.__blank_line_count>>" + str(self.__blank_line_count))
        self.__last_blank_line = token if token.is_blank_line else None

    def perform_close_check(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        """
        Perform any state checks necessary upon closing the heading context.  Also
        called at the end of a document to make sure the implicit close of the
        document is handled properly.
        """

        is_eligible_blank_line = token.is_blank_line
        is_simple_blank_line = True
        if is_eligible_blank_line and self.__last_blank_line is not None:
            line_number_delta = token.line_number - self.__last_blank_line.line_number
            is_simple_blank_line = line_number_delta == 1
            is_eligible_blank_line = is_simple_blank_line

        if (self.__start_heading_token and self.__did_heading_end) and (
            (not is_eligible_blank_line and not token.is_block_quote_end)
        ):
            did_end_match = self.__blank_line_count == self.__lines_below
            # print(">>END: did_end_match>>" + str(did_end_match))
            self.report_any_match_failures(context, did_end_match)
            self.__start_heading_token = None

        if token.is_blank_line and not is_simple_blank_line:
            self.__blank_line_count = 0

    def report_any_match_failures(
        self, context: PluginScanContext, did_end_match: bool
    ) -> None:
        """
        Take care of reporting any match failures.
        """

        assert self.__start_heading_token is not None
        if not self.__did_above_line_count_match:
            extra_info = f"Expected: {self.__lines_above}; Actual: {self.__start_heading_blank_line_count}; Above"
            # print(">>above>>" + extra_info)
            self.report_next_token_error(
                context,
                self.__start_heading_token,
                extra_error_information=extra_info,
                use_original_position=self.__start_heading_token.is_setext_heading,
            )
        if not did_end_match:
            extra_info = f"Expected: {self.__lines_below}; Actual: {self.__blank_line_count}; Below"
            # print(">>below>>" + extra_info)
            self.report_next_token_error(
                context,
                self.__start_heading_token,
                extra_error_information=extra_info,
                use_original_position=self.__start_heading_token.is_setext_heading,
            )


# pylint: enable=too-many-instance-attributes



"""
Module to provide for an encapsulation of the inline hard line break element.
"""

from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class HardBreakMarkdownToken(InlineMarkdownToken):
    """
    Class to provide for an encapsulation of the inline hard line break element.
    """

    def __init__(self, line_end: str, line_number: int, column_number: int) -> None:
        self.__line_end = line_end
        InlineMarkdownToken.__init__(
            self,
            MarkdownToken._token_inline_hard_break,
            MarkdownToken.extra_data_separator.join(
                [
                    self.__line_end,
                    ParserHelper.newline_character,
                ]
            ),
            line_number=line_number,
            column_number=column_number,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_inline_hard_break

    # pylint: enable=protected-access

    @property
    def line_end(self) -> str:
        """
        Returns the text at the end of the line.
        """
        return self.__line_end

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            HardBreakMarkdownToken, HardBreakMarkdownToken.__rehydrate_hard_break, None
        )

    @staticmethod
    def __rehydrate_hard_break(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the hard break text from the token.
        """
        _ = previous_token

        current_hard_break_token = cast(HardBreakMarkdownToken, current_token)
        leading_whitespace = (
            f"{current_hard_break_token.line_end}{ParserHelper.newline_character}"
        )
        if context.block_stack[-1].is_paragraph:
            block_paragraph_token = cast(
                ParagraphMarkdownToken, context.block_stack[-1]
            )
            (
                leading_whitespace,
                block_paragraph_token.rehydrate_index,
            ) = ParserHelper.recombine_string_with_whitespace(
                leading_whitespace,
                block_paragraph_token.extracted_whitespace,
                block_paragraph_token.rehydrate_index,
            )

        return "" if context.block_stack[-1].is_inline_link else leading_whitespace

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            HardBreakMarkdownToken,
            HardBreakMarkdownToken.__handle_hard_break_token,
            None,
        )

    @staticmethod
    def __handle_hard_break_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (next_token, transform_state)

        return "".join([output_html, "<br />", ParserHelper.newline_character])



"""
Module to implement a plugin that
"""

from typing import List, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.utils.container_token_manager import ContainerTokenManager
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RulePml101(RulePlugin):
    """
    Class to implement a plugin that
    """

    def __init__(self) -> None:
        super().__init__()
        self.__indent_basis = 0
        self.__container_manager = ContainerTokenManager()

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="list-anchored-indent",
            plugin_id="PML101",
            plugin_enabled_by_default=False,
            plugin_description="Anchored list indentation",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_pml101.md",
            plugin_configuration="indent",
            plugin_supports_fix=False,
            plugin_fix_level=3,
        )

    @classmethod
    def __validate_configuration_indent(cls, found_value: int) -> None:
        if found_value < 3 or found_value > 5:
            raise ValueError(
                "Allowable values are between 3 and 5."
            )  # pragma: no cover

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__indent_basis = (
            self.plugin_configuration.get_integer_property_with_default(
                "indent",
                4,
                valid_value_fn=self.__validate_configuration_indent,
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("indent", self.__indent_basis),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__container_manager.clear()

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_list_start or token.is_new_list_item:
            self.__check(context, token)

        self.__container_manager.manage_container_tokens(token)

    def __calculate_base_column_block_quote(
        self, stack_index: int, container_base_column: int, block_quote_base: int
    ) -> Tuple[int, int]:
        block_quote_token = cast(
            BlockQuoteMarkdownToken,
            self.__container_manager.container_token_stack[stack_index],
        )
        bq_index = self.__container_manager.bq_line_index[stack_index + 1]
        assert block_quote_token.bleading_spaces is not None
        split_leading_spaces = block_quote_token.bleading_spaces.split(
            ParserHelper.newline_character
        )
        if not block_quote_base:
            block_quote_base = container_base_column + len(
                split_leading_spaces[bq_index]
            )
        container_base_column += len(split_leading_spaces[bq_index])
        return container_base_column, block_quote_base

    def __calculate_base_column(self) -> Tuple[int, int, int]:
        container_base_column = 0
        block_quote_base = 0
        list_depth = 0
        if self.__container_manager.container_token_stack:
            stack_index = len(self.__container_manager.container_token_stack) - 1
            while (
                stack_index >= 0
                and self.__container_manager.container_token_stack[
                    stack_index
                ].is_list_start
            ):
                list_depth += 1
                stack_index -= 1
            while stack_index >= 0:
                if self.__container_manager.container_token_stack[
                    stack_index
                ].is_block_quote_start:
                    (
                        container_base_column,
                        block_quote_base,
                    ) = self.__calculate_base_column_block_quote(
                        stack_index, container_base_column, block_quote_base
                    )
                stack_index -= 1
        return container_base_column, block_quote_base, list_depth

    # def __check_apply_fix(
    #     self,
    #     context: PluginScanContext,
    #     token: MarkdownToken,
    #     adjusted_column_number: int,
    #     calculated_column_number: int,
    # ) -> None:
    #     list_token = cast(ListStartMarkdownToken, token)

    #     # column_delta is the space before the list start and follow_space_delta
    #     # is the space following the list start, before the list text commences.
    #     column_delta = adjusted_column_number - calculated_column_number
    #     follow_space_delta = (list_token.indent_level - list_token.column_number) - 1

    #     assert len(list_token.extracted_whitespace) >= column_delta
    #     adjusted_whitespace = list_token.extracted_whitespace[:-column_delta]

    #     self.register_fix_token_request(
    #         context,
    #         token,
    #         "next_token",
    #         "extracted_whitespace",
    #         adjusted_whitespace,
    #     )
    #     self.register_fix_token_request(
    #         context,
    #         token,
    #         "next_token",
    #         "indent_level",
    #         list_token.indent_level - column_delta - follow_space_delta,
    #     )
    #     if not token.is_new_list_item:
    #         self.register_fix_token_request(
    #             context,
    #             token,
    #             "next_token",
    #             "column_number",
    #             list_token.column_number - column_delta,
    #         )
    #         if list_token.leading_spaces:
    #             new_spaces = []
    #             total_delta = column_delta + follow_space_delta
    #             for i in list_token.leading_spaces.split("\n"):
    #                 if len(i) >= list_token.indent_level:
    #                     i = i[:-total_delta]
    #                 new_spaces.append(i)
    #             self.register_fix_token_request(
    #                 context,
    #                 list_token,
    #                 "next_token",
    #                 "leading_spaces",
    #                 "\n".join(new_spaces),
    #             )

    def __check(self, context: PluginScanContext, token: MarkdownToken) -> None:

        (
            container_base_column,
            block_quote_base,
            list_depth,
        ) = self.__calculate_base_column()

        if token.is_new_list_item:
            list_depth -= 1

        adjusted_column_number = token.column_number - 1 - container_base_column
        calculated_column_number = list_depth * self.__indent_basis
        if adjusted_column_number != calculated_column_number:
            # if context.in_fix_mode:
            #     self.__check_apply_fix(
            #         context, token, adjusted_column_number, calculated_column_number
            #     )
            # else:
            if block_quote_base:
                container_base_column -= block_quote_base
            # elif container_base_column:
            #     container_base_column += 1
            extra_error_information = (
                f"Expected: {calculated_column_number + container_base_column}, "
                + f"Actual={adjusted_column_number + container_base_column}"
            )
            self.report_next_token_error(
                context, token, extra_error_information=extra_error_information
            )



from typing import Optional

from pymarkdown.tokens.markdown_token import MarkdownToken, MarkdownTokenClass
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class SpecialMarkdownToken(MarkdownToken):
    def __init__(
        self,
        token_name: str,
        extra_data: Optional[str],
        line_number: int = 0,
        is_special: bool = True,
    ):
        MarkdownToken.__init__(
            self,
            token_name,
            MarkdownTokenClass.SPECIAL,
            extra_data,
            line_number=line_number,
            is_special=is_special,
        )


class EndOfStreamToken(SpecialMarkdownToken):
    def __init__(self, line_number: int) -> None:
        SpecialMarkdownToken.__init__(
            self, MarkdownToken._token_end_of_stream, None, line_number=line_number
        )

    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_end_of_stream

    def register_for_markdown_transform(
        self,
        registration_function: RegisterMarkdownTransformHandlersProtocol,
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """

        registration_function(
            EndOfStreamToken,
            EndOfStreamToken.__rehydrate_end_of_stream,
            None,
        )

    @staticmethod
    def __rehydrate_end_of_stream(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate nothing, as this is a placeholder.
        """
        _ = context, current_token, previous_token
        return ""

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            EndOfStreamToken,
            EndOfStreamToken.__handle_end_of_stream_token,
            None,
        )

    @staticmethod
    def __handle_end_of_stream_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return output_html



"""
Module to control the return code of the application.
"""

import sys
import threading
from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from enum import Enum
from typing import Dict

from application_properties import ApplicationProperties
from typing_extensions import override


class ApplicationResult(Enum):
    """
    Enumeration of all possible application results.
    """

    SUCCESS = 0
    NO_FILES_TO_SCAN = 1
    COMMAND_LINE_ERROR = 2
    FIXED_AT_LEAST_ONE_FILE = 3
    SCAN_TRIGGERED_AT_LEAST_ONCE = 4
    SYSTEM_ERROR = 5


class SchemeDefinition(ABC):
    """
    Abstract class for the scheme definitions.
    """

    @abstractmethod
    def get_scheme_mapping(self) -> Dict[ApplicationResult, int]:
        """
        Get the mapping to apply.
        """

    def apply_scheme(self, application_result: ApplicationResult) -> int:
        """
        Apply the scheme to the result to get the return code.
        """
        scheme_mapping = self.get_scheme_mapping()
        return scheme_mapping[application_result]


class DefaultScheme(SchemeDefinition):
    """
    Class to contain the default return code scheme.
    """

    @override
    def get_scheme_mapping(self) -> Dict[ApplicationResult, int]:
        return {
            ApplicationResult.SUCCESS: 0,
            ApplicationResult.NO_FILES_TO_SCAN: 1,
            ApplicationResult.COMMAND_LINE_ERROR: 2,
            ApplicationResult.FIXED_AT_LEAST_ONE_FILE: 3,
            ApplicationResult.SCAN_TRIGGERED_AT_LEAST_ONCE: 1,
            ApplicationResult.SYSTEM_ERROR: 1,
        }


class MinimalScheme(SchemeDefinition):
    """
    Class to contain the minimal return code scheme.
    """

    @override
    def get_scheme_mapping(self) -> Dict[ApplicationResult, int]:
        return {
            ApplicationResult.SUCCESS: 0,
            ApplicationResult.NO_FILES_TO_SCAN: 0,
            ApplicationResult.COMMAND_LINE_ERROR: 2,
            ApplicationResult.FIXED_AT_LEAST_ONE_FILE: 0,
            ApplicationResult.SCAN_TRIGGERED_AT_LEAST_ONCE: 0,
            ApplicationResult.SYSTEM_ERROR: 1,
        }


class ReturnCodeHelper:
    """
    Class to control the return code of the application.
    """

    __DEFAULT_SCHEME_NAME = "default"
    __MINIMAL_SCHEME_NAME = "minimal"

    __available_schemes: Dict[str, SchemeDefinition] = {
        __DEFAULT_SCHEME_NAME: DefaultScheme(),
        __MINIMAL_SCHEME_NAME: MinimalScheme(),
    }
    __helper_name = threading.local()

    @staticmethod
    def add_command_line_arguments(parser: ArgumentParser) -> None:
        """
        Function to add any command line arguments for this to the application.
        """
        parser.add_argument(
            "--return-code-scheme",
            dest="return_code_scheme",
            action="store",
            help="scheme to choose for selecting the application return code",
            type=ReturnCodeHelper.__validate_return_code_scheme,
            choices=list(ReturnCodeHelper.__available_schemes.keys()),
        )

    @staticmethod
    def reset() -> None:
        """
        Reset any mappings to their initial state.  Used to enforce strict mapping
        rules when called from within the same process.
        """
        ReturnCodeHelper.__helper_name.value = None

    @staticmethod
    def set_initial_state(args: Namespace, properties: ApplicationProperties) -> None:
        """
        Set the initial state, including the scheme for the return codes.
        """
        scheme_to_use = args.return_code_scheme
        if scheme_to_use is None:
            scheme_to_use = properties.get_string_property(
                "mode.return_code_scheme",
                strict_mode=True,
                valid_value_fn=ReturnCodeHelper.__validate_return_code_scheme,
            )
        if scheme_to_use is None:
            scheme_to_use = ReturnCodeHelper.__DEFAULT_SCHEME_NAME
        ReturnCodeHelper.__helper_name.value = scheme_to_use

    @staticmethod
    def exit_application(application_result: ApplicationResult) -> None:
        """
        Translate the ApplicationResult and exit the application.
        """
        scheme_name = (
            ReturnCodeHelper.__helper_name.value
            or ReturnCodeHelper.__DEFAULT_SCHEME_NAME
        )
        scheme_class = ReturnCodeHelper.__available_schemes[scheme_name]
        return_code = scheme_class.apply_scheme(application_result)
        sys.exit(return_code)

    @staticmethod
    def __validate_return_code_scheme(argument: str) -> str:
        """
        Function to help argparse limit the valid return code schemes.
        """
        if argument in ReturnCodeHelper.__available_schemes:
            return argument
        raise ValueError(f"Value '{argument}' is not a valid return code scheme.")



"""
Module to implement a plugin that ensures that a mandates set of headers are present.
"""

from typing import List, Optional, Tuple, Union, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd043(RulePlugin):
    """
    Class to implement a plugin that ensures that a mandates set of headers are present.
    """

    def __init__(self) -> None:
        super().__init__()
        # self.__show_debug = False
        self.__collected_tokens: List[MarkdownToken] = []
        self.__all_tokens: List[List[MarkdownToken]] = []
        self.__headings_have_wildcards: bool = False
        self.__compiled_headings: List[Union[str, Tuple[int, str]]] = []
        self.__prefix_index: int = -1
        self.__suffix_index: int = -1
        self.__raw_headings: str = ""

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="required-headings,required-headers",
            plugin_id="MD043",
            plugin_enabled_by_default=True,
            plugin_description="Required heading structure",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md043.md",
            plugin_configuration="headings",
        )

    @classmethod
    def __validate_heading_pattern(cls, found_value: str) -> None:
        if found_value.strip(" "):
            _, _, compile_error = cls.__compile(found_value)
            if compile_error:
                raise ValueError(f"Heading format not valid: {compile_error}")

    @classmethod
    def __compile(
        cls, found_value: str
    ) -> Tuple[List[Union[str, Tuple[int, str]]], bool, Optional[str]]:
        found_parts = found_value.split(",")
        compiled_lines: List[Union[str, Tuple[int, str]]] = []
        are_any_wildcards = False
        for next_part in found_parts:
            next_part = next_part.strip(" ")
            if next_part == "*":
                if compiled_lines and compiled_lines[-1] == "*":
                    return (
                        [],
                        False,
                        "Two wildcard elements cannot be next to each other.",
                    )
                compiled_lines.append(next_part)
                are_any_wildcards = True
            else:
                count, new_index = ParserHelper.collect_while_character(
                    next_part, 0, "#"
                )
                if not count:
                    return [], False, "Element must start with hash characters (#)."
                if count > 6:
                    return (
                        [],
                        False,
                        "Element must start with between 1 and 6 hash characters (#).",
                    )
                assert next_part is not None
                assert new_index is not None
                new_index, extracted_whitespace = ParserHelper.extract_ascii_whitespace(
                    next_part, new_index
                )
                if not extracted_whitespace or len(extracted_whitespace) != 1:
                    return (
                        [],
                        False,
                        "Element must have exactly one space character and one non-space character after any hash characters (#).",
                    )
                compiled_lines.append((count, next_part[new_index:]))
        return compiled_lines, are_any_wildcards, None

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        raw_headings = self.plugin_configuration.get_string_property_with_default(
            "headings",
            "",
            valid_value_fn=self.__validate_heading_pattern,
        )
        self.__raw_headings = raw_headings
        self.__compiled_headings = []
        self.__headings_have_wildcards = False
        if raw_headings:
            (
                self.__compiled_headings,
                self.__headings_have_wildcards,
                _,
            ) = self.__compile(raw_headings)
        if self.__headings_have_wildcards:
            self.__prefix_index = -1
            heading_index = 0
            while (
                heading_index < len(self.__compiled_headings)
                and self.__prefix_index == -1
            ):
                heading_contents = self.__compiled_headings[heading_index]
                if heading_contents == "*":
                    self.__prefix_index = heading_index
                else:
                    heading_index += 1

            self.__suffix_index = -1
            heading_index = len(self.__compiled_headings) - 1
            while heading_index >= 0 and self.__suffix_index == -1:
                heading_contents = self.__compiled_headings[heading_index]
                if heading_contents == "*":
                    self.__suffix_index = heading_index + 1
                else:
                    heading_index -= 1

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("headings", self.__raw_headings),
        ]

    def __verify_single_heading_match_atx(
        self, matching_all_token_index: int, hash_count: int, expected_text: str
    ) -> Tuple[Optional[MarkdownToken], Optional[str]]:
        failure_token, failure_reason = (None, None)

        these_tokens = self.__all_tokens[matching_all_token_index]
        atx_token = cast(AtxHeadingMarkdownToken, these_tokens[0])
        if atx_token.hash_count != hash_count:
            failure_reason = f"Bad heading level: Expected: {hash_count}, Actual: {atx_token.hash_count}"
        elif len(these_tokens) != 2 or not these_tokens[1].is_text:
            failure_reason = "Bad heading: Required headings must only be normal text."
        else:
            text_token = cast(TextMarkdownToken, these_tokens[1])
            if text_token.token_text != expected_text:
                failure_reason = f"Bad heading text: Expected: {expected_text}, Actual: {text_token.token_text}"
        if failure_reason:
            failure_token = these_tokens[0]
        return failure_token, failure_reason

    def __verify_single_heading_match(
        self,
        this_compiled_heading: Union[str, Tuple[int, str]],
        matching_all_token_index: int,
    ) -> Tuple[Optional[MarkdownToken], Optional[str]]:
        failure_token, failure_reason = (None, None)

        assert len(this_compiled_heading) == 2
        this_heading = cast(Tuple[int, str], this_compiled_heading)
        hash_count: int = this_heading[0]
        expected_text: str = this_heading[1]
        if matching_all_token_index >= len(self.__all_tokens):
            failure_token = self.__all_tokens[-1][0]
            failure_reason = f"Missing heading: {ParserHelper.repeat_string('#', hash_count)} {expected_text}"
        else:
            failure_token, failure_reason = self.__verify_single_heading_match_atx(
                matching_all_token_index, hash_count, expected_text
            )
        return failure_token, failure_reason

    def __verify_group_heading_match(
        self, heading_index: int, all_token_index: int, scan_limit: int = -1
    ) -> Tuple[int, int, Optional[MarkdownToken], Optional[str]]:
        if scan_limit < 0:
            scan_limit = len(self.__compiled_headings)

        failure_token, failure_reason = (None, None)
        # if self.__show_debug:
        #     print(
        #         "vghm:heading_index="
        #         + str(heading_index)
        #         + ",len="
        #         + str(len(self.__compiled_headings))
        #         + ",scan_limit="
        #         + str(scan_limit)
        #     )
        while (
            not failure_token
            and heading_index < len(self.__compiled_headings)
            and heading_index < scan_limit
        ):
            this_compiled_heading: Union[str, Tuple[int, str]] = (
                self.__compiled_headings[heading_index]
            )
            # if self.__show_debug:
            #     print(
            #         "vghm:this_compiled_heading="
            #         + str(this_compiled_heading)
            #         + ",all_token_index="
            #         + str(all_token_index)
            #     )
            failure_token, failure_reason = self.__verify_single_heading_match(
                this_compiled_heading, all_token_index
            )
            # if self.__show_debug:
            #     print(
            #         "vghm:failure_token="
            #         + str(failure_token)
            #         + ",failure_reason="
            #         + str(failure_reason)
            #     )
            if not failure_token:
                heading_index += 1
                all_token_index += 1
        return heading_index, all_token_index, failure_token, failure_reason

    def __handle_no_wildcards_match(self, context: PluginScanContext) -> None:
        end_index, _, failure_token, failure_reason = self.__verify_group_heading_match(
            0, 0
        )
        if failure_token:
            self.report_next_token_error(
                context,
                failure_token,
                extra_error_information=failure_reason,
                use_original_position=failure_token.is_setext_heading,
            )
        elif end_index < len(self.__all_tokens):
            anchor_token = self.__all_tokens[end_index][0]
            self.report_next_token_error(
                context,
                anchor_token,
                extra_error_information="Extra heading",
                use_original_position=anchor_token.is_setext_heading,
            )

    def __handle_wildcard_prefix(
        self,
    ) -> Tuple[int, int, Optional[MarkdownToken], Optional[str]]:
        (
            top_heading_index,
            top_token_index,
            failure_token,
            failure_reason,
        ) = self.__verify_group_heading_match(0, 0, scan_limit=self.__prefix_index)
        # if self.__show_debug:
        #     print(
        #         "top_heading_index="
        #         + str(top_heading_index)
        #         + ",top_token_index="
        #         + str(top_token_index)
        #         + ",failed_xx="
        #         + str(failure_reason)
        #     )
        return top_heading_index, top_token_index, failure_token, failure_reason

    def __handle_wildcard_suffix(
        self, top_token_index: int
    ) -> Tuple[int, int, Optional[MarkdownToken], Optional[str]]:
        bottom_heading_index = len(self.__compiled_headings)
        bottom_token_index = len(self.__all_tokens)

        start_all_token_index = len(self.__all_tokens) - (
            len(self.__compiled_headings) - self.__suffix_index
        )
        # if self.__show_debug:
        #     print(
        #         "start_all_token_index="
        #         + str(start_all_token_index)
        #         + ",top_token_index="
        #         + str(top_token_index)
        #     )
        if start_all_token_index < top_token_index:
            failure_token: Optional[MarkdownToken] = self.__all_tokens[-1][0]
            failure_reason: Optional[str] = "Overlapped."
        else:
            (
                _,
                _,
                failure_token,
                failure_reason,
            ) = self.__verify_group_heading_match(
                self.__suffix_index, start_all_token_index
            )
            # if self.__show_debug:
            #     print(
            #         "bottom_suffix_index="
            #         + str(bottom_suffix_index)
            #         + ",bottom_all_token_index="
            #         + str(bottom_all_token_index)
            #         + ",failure_reason="
            #         + str(failure_reason)
            #     )
            bottom_heading_index = self.__suffix_index
            bottom_token_index = start_all_token_index
        return bottom_heading_index, bottom_token_index, failure_token, failure_reason

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__collected_tokens = []
        self.__all_tokens = []

    def __do_recursive(
        self,
        remaining_headings: List[Union[str, Tuple[int, str]]],
        top_heading_index: int,
        remaining_tokens: List[List[MarkdownToken]],
        top_token_index: int,
    ) -> bool:
        bottom_heading_index = top_heading_index + len(remaining_headings)

        # if self.__show_debug:
        #     print(
        #         "remaining_headings:"
        #         + str(remaining_headings)
        #         + ",top="
        #         + str(top_heading_index)
        #         + ",bottom="
        #         + str(bottom_heading_index)
        #     )
        #     print(
        #         "remaining_tokens:"
        #         + str(remaining_tokens)
        #         + ",index="
        #         + str(top_token_index)
        #     )
        assert remaining_headings[0] == "*" and remaining_headings[-1] == "*"
        start_index = 1
        end_index = 2
        while remaining_headings[end_index] != "*":
            end_index += 1
        delta = end_index - start_index
        # if self.__show_debug:
        #     print(
        #         "start_index="
        #         + str(start_index)
        #         + ",end_index="
        #         + str(end_index)
        #         + ",delta="
        #         + str(delta)
        #     )
        #     print("headings:" + str(remaining_headings[start_index:end_index]))
        search_index = 0
        heading_index = top_heading_index + 1
        scan_limit = top_heading_index + 1 + delta
        found_match = False
        while search_index < len(remaining_headings) and search_index < len(
            remaining_tokens
        ):
            found_match, search_index = self.__do_recursive_loop(
                heading_index,
                top_token_index,
                search_index,
                scan_limit,
                bottom_heading_index,
                top_heading_index,
                remaining_tokens,
                remaining_headings,
            )
            if found_match:
                break
        return found_match

    # pylint: disable=too-many-arguments
    def __do_recursive_loop(
        self,
        heading_index: int,
        top_token_index: int,
        search_index: int,
        scan_limit: int,
        bottom_heading_index: int,
        top_heading_index: int,
        remaining_tokens: List[List[MarkdownToken]],
        remaining_headings: List[Union[str, Tuple[int, str]]],
    ) -> Tuple[bool, int]:
        (
            end_heading_index,
            end_token_index,
            failure_token,
            _,
        ) = self.__verify_group_heading_match(
            heading_index, top_token_index + search_index, scan_limit=scan_limit
        )
        # if self.__show_debug:
        #     print(str((end_heading_index, end_token_index, failure_token)))
        if not failure_token and end_heading_index < len(self.__compiled_headings):
            # if self.__show_debug:
            #     print(
            #         "FOUND:"
            #         + str((end_heading_index, end_token_index, failure_token))
            #     )
            if end_heading_index == bottom_heading_index - 1:
                # if self.__show_debug:
                #     print("done")
                found_match = True
            else:
                found_match = self.__prepare_and_do_recurse(
                    end_heading_index,
                    top_heading_index,
                    end_token_index,
                    top_token_index,
                    remaining_headings,
                    remaining_tokens,
                )
                # if self.__show_debug:
                #     print("found_match=" + str(found_match))
        else:
            found_match = False
        if not found_match:
            search_index += 1
        return found_match, search_index

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def __prepare_and_do_recurse(
        self,
        end_heading_index: int,
        top_heading_index: int,
        end_token_index: int,
        top_token_index: int,
        remaining_headings: List[Union[str, Tuple[int, str]]],
        remaining_tokens: List[List[MarkdownToken]],
    ) -> bool:
        new_top_heading_index = end_heading_index
        new_remaining_headings = remaining_headings[
            (end_heading_index - top_heading_index) :
        ]

        new_remaining_tokens = remaining_tokens[(end_token_index - top_token_index) :]
        new_top_token_index = end_token_index

        return self.__do_recursive(
            new_remaining_headings,
            new_top_heading_index,
            new_remaining_tokens,
            new_top_token_index,
        )

    # pylint: enable=too-many-arguments

    def completed_file(self, context: PluginScanContext) -> None:
        """
        Event that the file being currently scanned is now completed.
        """
        if not self.__compiled_headings:
            return

        # if self.__show_debug:
        #     print("self.__compiled_headings>" + str(self.__compiled_headings))
        #     print("self.__all_tokens>" + str(self.__all_tokens))
        failure_token = None

        if not self.__headings_have_wildcards:
            self.__handle_no_wildcards_match(context)
            return

        # if self.__show_debug:
        #     print(
        #         "__prefix_index="
        #         + str(self.__prefix_index)
        #         + ",__suffix_index="
        #         + str(self.__suffix_index)
        #         + ",len="
        #         + str(len(self.__compiled_headings))
        #     )
        top_heading_index = 0
        top_token_index = 0
        bottom_heading_index = len(self.__compiled_headings)
        bottom_token_index = len(self.__all_tokens)
        if self.__prefix_index != 0:
            (
                top_heading_index,
                top_token_index,
                failure_token,
                _,
            ) = self.__handle_wildcard_prefix()
        if not failure_token and self.__suffix_index != len(self.__compiled_headings):
            (
                bottom_heading_index,
                bottom_token_index,
                failure_token,
                _,
            ) = self.__handle_wildcard_suffix(top_token_index)
        if not failure_token:
            remaining_headings = self.__compiled_headings[
                top_heading_index:bottom_heading_index
            ]
            remaining_tokens = self.__all_tokens[top_token_index:bottom_token_index]
            # if self.__show_debug:
            #     print(
            #         "top_index="
            #         + str(top_heading_index)
            #         + ",bottom_index="
            #         + str(bottom_heading_index)
            #         + ",remaining_headings="
            #         + str(remaining_headings)
            #     )
            #     print(
            #         "top_token_index="
            #         + str(top_token_index)
            #         + ",bottom_token_index="
            #         + str(bottom_token_index)
            #         + ",remaining_tokens="
            #         + str(remaining_tokens)
            #     )
            if len(remaining_headings) != 1:
                recurse_result = self.__do_recursive(
                    remaining_headings,
                    top_heading_index,
                    remaining_tokens,
                    top_token_index,
                )
                if not recurse_result:
                    self.report_next_token_error(
                        context,
                        remaining_tokens[0][0],
                        extra_error_information="Multiple wildcard matching failed.",
                        use_original_position=remaining_tokens[0][0].is_setext_heading,
                    )
        if failure_token:
            self.report_next_token_error(
                context,
                failure_token,
                extra_error_information="Wildcard heading match failed.",
                use_original_position=failure_token.is_setext_heading,
            )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        _ = context

        if self.__collected_tokens:
            if token.is_atx_heading_end or token.is_setext_heading_end:
                self.__all_tokens.append(self.__collected_tokens)
                self.__collected_tokens = []
            else:
                self.__collected_tokens.append(token)
        elif self.__compiled_headings:
            if token.is_atx_heading or token.is_setext_heading:
                self.__collected_tokens = [token]



"""
Module to implement a plugin that looks for image elements that do not specify alternate text.
"""

from typing import cast

from pymarkdown.general.constants import Constants
from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.image_start_markdown_token import ImageStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd045(RulePlugin):
    """
    Class to implement a plugin that looks for image elements that do not specify alternate text.
    """

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-alt-text",
            plugin_id="MD045",
            plugin_enabled_by_default=True,
            plugin_description="Images should have alternate text (alt text)",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md045.md",
        )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_inline_image:
            image_token = cast(ImageStartMarkdownToken, token)
            if not image_token.text_from_blocks.strip(
                Constants.unicode_whitespace.value()
            ):
                self.report_next_token_error(context, token)



"""
Module to provide for disallowing raw HTML in document.
"""

import string
from typing import Set

from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.extension_manager.extension_manager_constants import (
    ExtensionManagerConstants,
)
from pymarkdown.extension_manager.parser_extension import ParserExtension
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade


class MarkdownDisallowRawHtmlExtension(ParserExtension):
    """
    Extension to implement the disallow rawhtml extension.
    """

    def __init__(self) -> None:
        self.__disallowed_tag_names: Set[str] = set()

    def get_identifier(self) -> str:
        """
        Get the identifier associated with this extension.
        """
        return "markdown-disallow-raw-html"

    def get_details(self) -> ExtensionDetails:
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id=self.get_identifier(),
            extension_name="Markdown Disallow Raw HTML",
            extension_description="Disallows parsing of any raw HTML.",
            extension_enabled_by_default=False,
            extension_version="0.5.0",
            extension_interface_version=ExtensionManagerConstants.EXTENSION_INTERFACE_VERSION_BASIC,
            extension_url="https://pymarkdown.readthedocs.io/en/latest/extensions/disallowed-raw-html/",
            extension_configuration=None,
        )

    __valid_tag_name_characters = f"{string.ascii_letters}{string.digits}-"

    @staticmethod
    def is_valid_tag_name(tag_name: str) -> bool:
        """
        Determine if the html tag name is valid according to the html rules.
        """

        return (
            all(
                next_character
                in MarkdownDisallowRawHtmlExtension.__valid_tag_name_characters
                for next_character in tag_name.lower()
            )
            if tag_name
            else False
        )

    def apply_configuration(
        self, extension_specific_facade: MyApplicationPropertiesFacade
    ) -> None:
        """
        Apply any configuration required by the extension.
        """
        self.__disallowed_tag_names = {
            "title",
            "textarea",
            "style",
            "xmp",
            "iframe",
            "noembed",
            "noframes",
            "script",
            "plaintext",
        }

        modify_tag_names = extension_specific_facade.get_string_property(
            "change_tag_names", default_value=None
        )
        if modify_tag_names is not None:
            tag_config_name = f"extensions.{self.get_identifier()}.change_tag_names"
            for next_tag_part in modify_tag_names.split(","):
                next_tag_part = next_tag_part.strip(" ")
                if not next_tag_part:
                    raise ValueError(
                        f"Configuration item '{tag_config_name}' contains at least one empty string."
                    )
                if next_tag_part[0] not in ("+", "-"):
                    raise ValueError(
                        f"Configuration item '{tag_config_name}' elements must either start with '+' or '-'."
                    )
                remaining_tag_part = next_tag_part[1:]
                if not MarkdownDisallowRawHtmlExtension.is_valid_tag_name(
                    remaining_tag_part
                ):
                    raise ValueError(
                        f"Configuration item '{tag_config_name}' contains an element '{remaining_tag_part}' that is not a valid tag name."
                    )
                if next_tag_part[0] == "+":
                    self.__disallowed_tag_names.add(remaining_tag_part.lower())
                else:
                    self.__disallowed_tag_names.remove(remaining_tag_part.lower())

    def is_html_tag_disallowed(self, tag_name: str) -> bool:
        """
        Determine if the specified tag name is allowed.
        """
        tag_name = tag_name.lower()
        return tag_name in self.__disallowed_tag_names



"""
Module to provide for the output of the PyMarkdown application.
"""

import sys
from typing import Optional

from pymarkdown.plugin_manager.plugin_scan_failure import PluginScanFailure


class MainPresentation:
    """
    Class to provide for the output of the PyMarkdown application.
    """

    def print_system_output(self, output_string: str) -> None:
        """
        Root function to output to standard out.
        """
        print(output_string, file=sys.stdout)

    def print_system_error(self, error_string: str) -> None:
        """
        Root function to output to standard error.
        """
        print(error_string, file=sys.stderr)

    def format_scan_error(
        self,
        next_file: str,
        this_exception: Exception,
        show_extended_information: bool = False,
        allow_shortcut: bool = False,
    ) -> Optional[str]:
        """
        Format a scan error for display.  Returning a value of None means that
        the function has handled any required output.
        """
        if allow_shortcut:
            formatted_error = f"{next_file}:0:0: {this_exception}"
        else:
            formatted_error = f"{type(this_exception).__name__} encountered while scanning '{next_file}':\n{this_exception}"
            if show_extended_information:
                current_cause = this_exception.__cause__
                while current_cause:
                    formatted_error += f"\nCaused by: {type(current_cause).__name__}:\n   {current_cause}"
                    current_cause = current_cause.__cause__
        return formatted_error

    def print_pragma_failure(
        self, scan_file: str, line_number: int, pragma_error: str
    ) -> None:
        """
        Print a failure to compile the pragma.
        """
        self.print_system_error(f"{scan_file}:{line_number}:1: INLINE: {pragma_error}")

    def print_scan_failure(self, scan_failure: PluginScanFailure) -> None:
        """
        Print a scan failure for a specific file and location.
        """
        self.print_system_output(
            f"{scan_failure.scan_file}:{scan_failure.line_number}:{scan_failure.column_number}: "
            + f"{scan_failure.rule_id}: {scan_failure.rule_description}{scan_failure.extra_error_information} ({scan_failure.rule_name})"
        )

    def print_fix_message(self, file_fixed: str) -> None:
        """
        Print a message indicating that a given file has been fixed.
        """
        self.print_system_output(f"Fixed: {file_fixed}")
